from __future__ import absolute_import

import numpy as np
import time
import copy
import os
import re
import math
from collections import OrderedDict
import multiprocessing
import multiprocessing.pool
import h5py
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import lmfit
import platform
from distutils.version import LooseVersion
import dask.array as da

from atom.api import Atom, Str, observe, Typed, Int, List, Dict, Float, Bool
from skbeam.core.fitting.xrf_model import (
    ModelSpectrum,
    update_parameter_dict,
    # sum_area,
    set_parameter_bound,
    # ParamController,
    K_LINE,
    L_LINE,
    M_LINE,
    nnls_fit,
    construct_linear_model,
    # linear_spectrum_fitting,
    register_strategy,
    TRANSITIONS_LOOKUP,
)
from skbeam.fluorescence import XrfElement as Element
from .parameters import calculate_profile, fit_strategy_list, trim_escape_peak, define_range
from .fileio import save_fitdata_to_hdf, output_data

from ..core.fitting import rfactor
from ..core.quant_analysis import ParamQuantEstimation
from ..core.map_processing import fit_xrf_map, TerminalProgressBar, prepare_xrf_map, snip_method_numba

import logging

logger = logging.getLogger(__name__)


class Fit1D(Atom):
    """
    Fit 1D fluorescence spectrum. Users can choose multiple strategies
    for this fitting.

    Parameters
    ----------
    x_data : 2D array
        x position data from hdf file
    y_data : 2D array
        y position data from hdf file
    result_map : Dict
        dict of 2D array map for each fitted element
    map_interpolation : bool
        option to interpolate the 2D map according to x,y position or not
        Interpolation is performed only before exporting data
        (currently .tiff or .txt)
    hdf_path : str
        path to hdf file
    hdf_name : str
        name of hdf file
    """

    file_status = Str()

    img_dict = Dict()

    element_list = List()
    fit_x = Typed(np.ndarray)
    fit_y = Typed(np.ndarray)
    residual = Typed(np.ndarray)
    comps = Dict()
    fit_strategy1 = Int(0)
    fit_strategy2 = Int(0)
    fit_strategy3 = Int(0)
    fit_strategy4 = Int(0)
    fit_strategy5 = Int(0)
    fit_result = Typed(object)
    data_title = Str()
    runid = Int(0)

    working_directory = Str()
    result_folder = Str()

    all_strategy = Typed(object)

    x0 = Typed(np.ndarray)
    y0 = Typed(np.ndarray)
    bg = Typed(np.ndarray)
    es_peak = Typed(np.ndarray)
    cal_x = Typed(np.ndarray)
    cal_y = Typed(np.ndarray)
    cal_spectrum = Dict()

    #  attributes used by the ElementEdit window
    selected_element = Str()
    selected_index = Int()
    elementinfo_list = List()

    # The variable contains the image title displayed in "Element Map' tab.
    #   The variable is synchronized with the identical variable in 'DrawImageAdvanced' class.
    img_title = Str()
    # Reference to the dictionary that contains dataset currently displayed in 'Element Map' tab
    #   The variable is synchronized with the identical variable in 'DrawImageAdvanced' class
    dict_to_plot = Dict()

    function_num = Int(0)
    nvar = Int(0)
    chi2 = Float(0.0)
    red_chi2 = Float(0.0)
    r2 = Float(0.0)
    global_param_list = List()

    fit_num = Int(100)
    ftol = Float(1e-5)
    c_weight = Float(1e2)

    fit_img = Dict()

    save_point = Bool(False)
    point1v = Int(0)
    point1h = Int(0)
    point2v = Int(0)
    point2h = Int(0)

    e_name = Str()
    add_element_intensity = Float(100.0)
    pileup_data = Dict()

    raise_bg = Float(0.0)
    pixel_bin = Int(0)
    linear_bg = Bool(False)
    use_snip = Bool(True)
    bin_energy = Int(0)
    fit_info = Str()
    pixel_fit_info = Str()

    pixel_fit_method = Int(0)

    result_map = Dict()
    map_interpolation = Bool(False)
    hdf_path = Str()
    hdf_name = Str()

    roi_sum_opt = Dict()
    scaler_keys = List()
    scaler_index = Int(0)

    # Reference to ParamModel object
    param_model = Typed(object)
    # Reference to FileIOModel object
    io_model = Typed(object)

    # Reference to FileIOMOdel
    io_model = Typed(object)

    # Quantitative analysis: used during estimation step
    param_quant_estimation = ParamQuantEstimation()
    # *** The following two references are used exclusively to update the list of standards ***
    # ***   in the dialog box. ***
    qe_param_built_in_ref = Typed(object)
    qe_param_custom_ref = Typed(object)
    # The following reference used to track the selected standard while the selection
    #   dialog box is open. Once the dialog box is opened again, the reference becomes
    #   invalid, since the descriptions of the standards are reloaded from files.
    qe_standard_selected_ref = Typed(object)
    # Keep the actual copy of the selected standard. The copy is used to keep information
    #   on the selected standard while descriptions are reloaded (references become invalid).
    qe_standard_selected_copy = Typed(object)
    # *** The following fields are used exclusively to store input values ***
    # ***   for the 'SaveQuantCalibration' dialog box. ***
    # *** The fields are not guaranteed to have valid values at any other time. ***
    qe_standard_distance_to_sample = Float(0.0)

    def __init__(self, *, param_model, io_model, working_directory):
        self.working_directory = working_directory
        self.result_folder = working_directory
        self.all_strategy = OrderedDict()

        # Reference to ParamModel object
        self.param_model = param_model

        # Reference to FileIOMOdel
        self.io_model = io_model

        self.pileup_data = {"element1": "Si_K", "element2": "Si_K", "intensity": 100.0}

        # don't argue, plotting purposes
        self.fit_strategy1 = 0
        self.fit_strategy2 = 0
        self.fit_strategy1 = 1
        self.fit_strategy2 = 0

        # perform roi sum in given range
        self.roi_sum_opt["status"] = False
        self.roi_sum_opt["low"] = 0.0
        self.roi_sum_opt["high"] = 10.0

        self.qe_standard_selected_ref = None
        self.qe_standard_selected_copy = None

    def result_folder_changed(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.result_folder = change["value"]

    def data_title_update(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.data_title = change["value"]

    def runid_update(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.runid = change["value"]

    def img_dict_updated(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : bool
            True - 'io_model.img_dict` was updated, False - ignore
        """
        if change["value"]:
            _key = [k for k in self.io_model.img_dict.keys() if "scaler" in k]
            if len(_key) != 0:
                self.scaler_keys = sorted(self.io_model.img_dict[_key[0]].keys())

    def scaler_index_update(self, change):
        """
        Observer function to be connected to the fit_model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.scaler_index = change["value"]

    def img_title_update(self, change):
        r"""Observer function. Sets ``img_title`` field to the same value as
        the identical variable in ``DrawImageAdvanced`` class"""
        self.img_title = change["value"]

    def dict_to_plot_update(self, change):
        r"""Observer function. Sets ``dict_to_plot`` field to the same value as
        the identical variable in ``DrawImageAdvanced`` class"""
        self.dict_to_plot = change["value"]

    def update_selected_index(self, selected_element=None, element_list_new=None):

        if selected_element is None:
            # Currently selected element
            element = self.selected_element
        else:
            # Selected element (probably saved before update of the element list)
            element = selected_element

        if element_list_new is None:
            # Current element list
            element_list = self.param_model.element_list
        else:
            # Future element list (before it is updated)
            element_list = element_list_new

        if not element_list:
            # Empty element list
            ind = 0
        else:
            try:
                # Combo-box has additional element 'Select element'
                ind = element_list.index(element) + 1
            except ValueError:
                # Element is not found (was deleted), so deselect the element
                ind = 0
        if ind == self.selected_index:
            # We want the content to update (deselect, then select again)
            self.selected_index = 0
        self.selected_index = ind

    @observe("selected_index")
    def _selected_element_changed(self, change):
        if change["value"] > 0:
            ind_sel = change["value"] - 1
            if ind_sel >= len(self.param_model.element_list):
                ind_sel = len(self.param_model.element_list) - 1
                self.selected_index = ind_sel + 1  # Change the selection as well
            self.selected_element = self.param_model.element_list[ind_sel]
            if len(self.selected_element) <= 4:
                element = self.selected_element.split("_")[0]
                self.elementinfo_list = sorted(
                    [
                        e
                        for e in list(self.param_model.param_new.keys())
                        if (element + "_" in e) and ("pileup" not in e)  # error between S_k or Si_k
                    ]
                )  # Si_ka1 not Si_K
                logger.info(f"Element line info: {self.elementinfo_list}")
            else:
                element = self.selected_element  # for pileup peaks
                self.elementinfo_list = sorted(
                    [e for e in list(self.param_model.param_new.keys()) if element.replace("-", "_") in e]
                )
                logger.info(f"User defined or pileup peak info: {self.elementinfo_list}")
        else:
            self.elementinfo_list = []

    def select_index_by_eline_name(self, eline_name):
        # Select the element by name. If element is selected, then the ``elementinfo_list`` with
        #   names of parameters is created. Originally the element could only be selected
        #   by its index in the list (from dialog box ``ElementEdit``. This function is
        #   used by interface components for editing parameters of user defined peaks.
        if eline_name in self.param_model.element_list:
            # This will fill the list ``self.elementinfo_list``
            self.selected_index = self.param_model.element_list.index(eline_name) + 1
        else:
            raise Exception(f"Line '{eline_name}' is not in the list of selected element lines.")

    def apply_default_param(self):
        """
        Update parameters with default parameters, also update element list.
        """
        # Save currently selected element name
        selected_element = self.selected_element
        self.selected_index = 0

        element_list = self.param_model.param_new["non_fitting_values"]["element_list"]
        element_list = [e.strip(" ") for e in element_list.split(",")]
        element_list = [_ for _ in element_list if _]  # Get rid of empty strings in the list
        self.param_model.element_list = element_list

        # Update 'self.param_model.EC'
        # self.param_model.create_spectrum_from_param_dict()

        self.update_element_info()

        # Update the index in case the selected emission line disappeared from the list
        self.update_selected_index(selected_element=selected_element, element_list_new=element_list)

        # global parameters
        # for GUI purpose only
        # if we do not clear the list first, there is not update on the GUI
        self.global_param_list = []
        self.global_param_list = sorted(
            [k for k in self.param_model.param_new.keys() if k == k.lower() and k != "non_fitting_values"]
        )

        self.define_range()

        #  register the strategy and extend the parameter list
        #  to cover all given elements
        # for strat_name in fit_strategy_list:
        #     strategy = extract_strategy(self.param_dict, strat_name)
        #     #  register the strategy and extend the parameter list
        #     #  to cover all given elements
        #     register_strategy(strat_name, strategy)
        #     set_parameter_bound(self.param_dict, strat_name)

        #  define element_adjust as fixed
        # self.param_dict = define_param_bound_type(self.param_dict)

    def filename_update(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : dict
            This is the dictionary that gets passed to a function
            with the @observe decorator
        """
        self.hdf_name = change["value"]
        # output to .h5 file
        self.hdf_path = os.path.join(self.result_folder, self.hdf_name)

    @observe("fit_strategy1")
    def update_strategy1(self, change):
        self.all_strategy.update({"strategy1": change["value"]})
        if change["value"]:
            logger.info("Setting strategy (preset) for step 1: {}".format(fit_strategy_list[change["value"] - 1]))

    @observe("fit_strategy2")
    def update_strategy2(self, change):
        self.all_strategy.update({"strategy2": change["value"]})
        if change["value"]:
            logger.info("Setting strategy (preset) for step 2: {}".format(fit_strategy_list[change["value"] - 1]))

    @observe("fit_strategy3")
    def update_strategy3(self, change):
        self.all_strategy.update({"strategy3": change["value"]})
        if change["value"]:
            logger.info("Setting strategy (preset) for step 3: {}".format(fit_strategy_list[change["value"] - 1]))

    @observe("fit_strategy4")
    def update_strategy4(self, change):
        self.all_strategy.update({"strategy4": change["value"]})
        if change["value"]:
            logger.info("Strategy at step 4 is: {}".format(fit_strategy_list[change["value"] - 1]))

    @observe("fit_strategy5")
    def update_strategy5(self, change):
        self.all_strategy.update({"strategy5": change["value"]})
        if change["value"]:
            logger.info("Strategy at step 5 is: {}".format(fit_strategy_list[change["value"] - 1]))

    def update_param_with_result(self):
        update_parameter_dict(self.param_model.param_new, self.fit_result)

    def define_range(self):
        """
        Cut x range according to values define in param_dict.
        """
        lowv = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        highv = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]
        self.x0, self.y0 = define_range(
            self.io_model.data,
            lowv,
            highv,
            self.param_model.param_new["e_offset"]["value"],
            self.param_model.param_new["e_linear"]["value"],
        )

    def get_background(self):
        self.bg = snip_method_numba(
            self.y0,
            self.param_model.param_new["e_offset"]["value"],
            self.param_model.param_new["e_linear"]["value"],
            self.param_model.param_new["e_quadratic"]["value"],
            width=self.param_model.param_new["non_fitting_values"]["background_width"],
        )

    def get_profile(self):
        """
        Calculate profile based on current parameters.
        """
        # self.define_range()

        #  Do nothing if no data is loaded
        if self.x0 is None or self.y0 is None:
            return

        self.cal_x, self.cal_spectrum, area_dict = calculate_profile(
            self.x0, self.y0, self.param_model.param_new, self.param_model.element_list
        )
        #  add escape peak
        if self.param_model.param_new["non_fitting_values"]["escape_ratio"] > 0:
            self.cal_spectrum["escape"] = trim_escape_peak(
                self.io_model.data, self.param_model.param_new, len(self.y0)
            )

        self.cal_y = np.zeros(len(self.cal_x))
        for k, v in self.cal_spectrum.items():
            self.cal_y += v

        self.residual = self.cal_y - self.y0

    def fit_data(self, x0, y0, *, init_params=True):
        fit_num = self.fit_num
        ftol = self.ftol
        c_weight = 1  # avoid zero point

        params = copy.deepcopy(self.param_model.param_new)
        # Initialize parameters to some large default value. This increases the rate of convergence
        #   and prevents optimization algorithm from getting stuck. There could be better solution,
        #   but this seems to be functional workaround.
        if init_params:
            for k, p in params.items():
                if "area" in k:
                    p["value"] = 1000

        MS = ModelSpectrum(params, self.param_model.element_list)
        MS.assemble_models()

        # weights = 1/(c_weight + np.abs(y0))
        weights = 1 / np.sqrt(c_weight + np.abs(y0))
        # weights /= np.sum(weights)
        result = MS.model_fit(x0, y0, weights=weights, maxfev=fit_num, xtol=ftol, ftol=ftol, gtol=ftol)
        self.fit_x = (
            result.values["e_offset"] + result.values["e_linear"] * x0 + result.values["e_quadratic"] * x0 ** 2
        )
        self.fit_y = result.best_fit
        self.fit_result = result
        self.residual = self.fit_y - y0

    def fit_multiple(self):
        """
        Fit data in sequence according to given strategies.
        The param_dict is extended to cover elemental parameters.
        Use app.precessEvents() for multi-threading.
        """
        # app = QApplication.instance()
        self.define_range()
        self.get_background()

        # PC = ParamController(self.param_dict, self.param_model.element_list)
        # self.param_dict = PC.params

        if self.param_model.param_new["non_fitting_values"]["escape_ratio"] > 0:
            self.es_peak = trim_escape_peak(self.io_model.data, self.param_model.param_new, self.y0.size)
            y0 = self.y0 - self.bg - self.es_peak
        else:
            y0 = self.y0 - self.bg

        t0 = time.time()
        self.fit_info = (
            "Spectrum fitting of the sum spectrum (incident energy "
            f"{self.param_model.param_new['coherent_sct_energy']['value']})."
        )
        # app.processEvents()
        # logger.info('-------- '+self.fit_info+' --------')

        # Parameters should be initialized only once
        init_params = True
        for k, v in self.all_strategy.items():
            if v:
                strat_name = fit_strategy_list[v - 1]
                # self.fit_info = 'Fit with {}: {}'.format(k, strat_name)

                logger.info(self.fit_info)
                strategy = extract_strategy(self.param_model.param_new, strat_name)
                #  register the strategy and extend the parameter list
                #  to cover all given elements
                register_strategy(strat_name, strategy)
                set_parameter_bound(self.param_model.param_new, strat_name)

                self.fit_data(self.x0, y0, init_params=init_params)
                init_params = False

                self.update_param_with_result()

                # The following is a patch for rare cases when fitting results in negative
                #   areas for some emission lines. These are typically non-existent lines, but
                #   they should not be automatically eliminated from the list. To prevent
                #   elimination, set the area to some small positive value.
                for key, val in self.param_model.param_new.items():
                    if key.endswith("_area") and val["value"] <= 0.0:
                        _small_value_for_area = 0.1
                        logger.warning(
                            f"Fitting resulted in negative value for '{key}' ({val['value']}). \n"
                            f"    In order to continue using the emission line in future computations, "
                            f"the fitted area is set to a small value ({_small_value_for_area}).\n    Delete "
                            f"the emission line from the list if you know it is not present in "
                            f"the sample."
                        )
                        val["value"] = _small_value_for_area  # Some small number

                #  calculate r2
                self.r2 = cal_r2(y0, self.fit_y)
                self.assign_fitting_result()
                # app.processEvents()

        t1 = time.time()
        logger.warning("Time used for summed spectrum fitting is : {}".format(t1 - t0))

        self.comps.clear()
        comps = self.fit_result.eval_components(x=self.x0)
        self.comps = combine_lines(comps, self.param_model.element_list, self.bg)

        if self.param_model.param_new["non_fitting_values"]["escape_ratio"] > 0:
            self.fit_y += self.bg + self.es_peak
            self.comps["escape"] = self.es_peak
        else:
            self.fit_y += self.bg

        self.save_result()
        self.assign_fitting_result()
        self.fit_info = "Summed spectrum fitting is done!"
        logger.info("-------- " + self.fit_info + " --------")

    def compute_current_rfactor(self, save_fit=True):
        """
        Compute current R-factor value. The fitted array is selected
        based on `save_fit`. The same arrays are selected as in `output_summed_data_fit`

        Parameters
        ----------
        save_fit: bool
            True - use total spectrum fitting data (available after fitting was done),
            False - use weighted spectral components with weights equal to current parameters.

        Returns
        -------
        float or None
            Value of R-factor or None if R-factor can not be computed.
        """
        rf = None
        # R-factor is for visualization purposes only, so display 0 if it can not be computed.
        if save_fit:
            if (self.y0 is not None) and (self.fit_y is not None) and (self.y0.shape == self.fit_y.shape):
                rf = rfactor(self.y0, self.fit_y)
        else:
            if (
                (self.param_model.y0 is not None)
                and (self.param_model.total_y is not None)
                and (self.param_model.y0.shape == self.param_model.total_y.shape)
            ):
                rf = rfactor(self.param_model.y0, self.param_model.total_y)
        return rf

    def output_summed_data_fit(self, save_fit=True):
        """Save energy, summed data and fitting curve to a file."""
        xx = None
        if self.x0 is not None:
            a0, a1, a2 = (
                self.param_model.param_new["e_offset"]["value"],
                self.param_model.param_new["e_linear"]["value"],
                self.param_model.param_new["e_quadratic"]["value"],
            )
            xx = a0 + self.x0 * a1 + self.x0 ** 2 * a2
        if save_fit:
            logger.info("Saving spectrum after total spectrum fitting.")
            if (xx is None) or (self.y0 is None) or (self.fit_y is None):
                msg = "Not enough data to save spectrum/fit data. Total spectrum fitting was not run."
                raise RuntimeError(msg)
            data = np.array([self.x0, self.y0, self.fit_y])
        else:
            logger.info("Saving spectrum based on loaded or estimated parameters.")
            if (xx is None) or (self.y0 is None) or (self.param_model.total_y is None):
                msg = "Not enough data to save spectrum/fit data based on loaded or estimated parameters."
                raise RuntimeError(msg)
            data = np.array([xx, self.y0, self.param_model.total_y])

        output_fit_name = self.data_title + "_summed_spectrum_fit.txt"
        fpath = os.path.join(self.result_folder, output_fit_name)
        np.savetxt(fpath, data.T)
        logger.info(f"Spectrum fit data is saved to file '{fpath}'")

    def assign_fitting_result(self):
        self.function_num = self.fit_result.nfev
        self.nvar = self.fit_result.nvarys
        # self.chi2 = np.around(self.fit_result.chisqr, 4)
        self.red_chi2 = np.around(self.fit_result.redchi, 4)

    def fit_single_pixel(self):
        """
        This function performs single pixel fitting.
        Multiprocess is considered.
        """
        # app = QApplication.instance()
        raise_bg = self.raise_bg
        pixel_bin = self.pixel_bin
        comp_elastic_combine = False
        linear_bg = self.linear_bg
        use_snip = self.use_snip
        bin_energy = self.bin_energy

        if self.pixel_fit_method == 0:
            pixel_fit = "nnls"
        elif self.pixel_fit_method == 1:
            pixel_fit = "nonlinear"

        logger.info(
            "-------- Fitting of single pixels starts (incident_energy "
            f"{self.param_model.param_new['coherent_sct_energy']['value']} keV) --------"
        )
        t0 = time.time()
        self.pixel_fit_info = "Pixel fitting is in process."

        # app.processEvents()
        self.result_map, calculation_info = single_pixel_fitting_controller(
            self.io_model.data_all,
            self.param_model.param_new,
            method=pixel_fit,
            pixel_bin=pixel_bin,
            raise_bg=raise_bg,
            comp_elastic_combine=comp_elastic_combine,
            linear_bg=linear_bg,
            use_snip=use_snip,
            bin_energy=bin_energy,
        )

        t1 = time.time()
        logger.info("Time used for pixel fitting is : {}".format(t1 - t0))

        #  get fitted spectrum and save them to figs
        if self.save_point is True:
            self.pixel_fit_info = "Saving output ..."
            # app.processEvents()
            elist = calculation_info["fit_name"]
            matv = calculation_info["regression_mat"]
            results = calculation_info["results"]
            # fit_range = calculation_info['fit_range']
            x = calculation_info["energy_axis"]
            x = (
                self.param_model.param_new["e_offset"]["value"]
                + self.param_model.param_new["e_linear"]["value"] * x
                + self.param_model.param_new["e_quadratic"]["value"] * x ** 2
            )
            data_fit = calculation_info["input_data"]
            data_sel_indices = calculation_info["data_sel_indices"]

            p1 = [self.point1v, self.point1h]
            p2 = [self.point2v, self.point2h]

            if self.point2v > 0 or self.point2h > 0:
                prefix_fname = self.hdf_name.split(".")[0]
                output_folder = os.path.join(self.result_folder, prefix_fname + "_pixel_fit")
                if os.path.exists(output_folder) is False:
                    os.mkdir(output_folder)
                save_fitted_fig(
                    x,
                    matv,
                    results[:, :, 0 : len(elist)],
                    p1,
                    p2,
                    data_fit,
                    data_sel_indices,
                    self.param_model.param_new,
                    output_folder,
                    use_snip=use_snip,
                )

            # the output movie are saved as the same name
            # need to define specific name for prefix, to be updated
            # save_fitted_as_movie(x, matv, results[:, :, 0:len(elist)],
            #                      p1, p2,
            #                      data_fit, self.param_dict,
            #                      self.result_folder, prefix=prefix_fname, use_snip=use_snip)
            logger.info("Done with saving fitting plots.")
        try:
            self.save2Dmap_to_hdf(calculation_info=calculation_info, pixel_fit=pixel_fit)
            self.pixel_fit_info = "Pixel fitting is done!"
            # app.processEvents()
        except ValueError:
            logger.warning("Fitting result can not be saved to h5 file.")
        except IOError as ex:
            logger.warning(f"{ex}")
        logger.info("-------- Fitting of single pixels is done! --------")

    def save_pixel_fitting_to_db(self):
        """Save fitting results to analysis store"""
        from .data_to_analysis_store import save_data_to_db

        doc = {}
        doc["param"] = self.param_model.param_new
        doc["exp"] = self.io_model.data
        doc["fitted"] = self.fit_y
        save_data_to_db(self.runid, self.result_map, doc)

    def save2Dmap_to_hdf(self, *, calculation_info=None, pixel_fit="nnls"):
        """
        Save fitted 2D map of elements into hdf file after fitting is done. User
        can choose to interpolate the image based on x,y position or not.

        Parameters
        ----------
        pixel_fit : str
            If nonlinear is chosen, more information needs to be saved.
        """

        prefix_fname = self.hdf_name.split(".")[0]
        if len(prefix_fname) == 0:
            prefix_fname = "tmp"

        # Generate the path to computed ROIs in the HDF5 file
        det_name = "detsum"  # Assume that ROIs are computed using the sum of channels
        fit_name = f"{prefix_fname}_fit"

        # Search for channel name in the data title. Channels are named
        #   det1, det2, ... , i.e. 'det' followed by integer number.
        # The channel name is always located at the end of the ``data_title``.
        # If the channel name is found, then build the path using this name.
        srch = re.search("det\d+$", self.data_title)  # noqa: W605
        if srch:
            det_name = srch.group(0)
            fit_name = f"{prefix_fname}_{det_name}_fit"
        inner_path = f"xrfmap/{det_name}"

        # Update GUI so that results can be seen immediately
        self.fit_img[fit_name] = self.result_map

        if not os.path.isfile(self.hdf_path):
            raise IOError(f"File '{self.hdf_path}' does not exist. Data is not saved to HDF5 file.")

        save_fitdata_to_hdf(self.hdf_path, self.result_map, datapath=inner_path)

        # output error
        if pixel_fit == "nonlinear":
            error_map = calculation_info["error_map"]
            save_fitdata_to_hdf(
                self.hdf_path,
                error_map,
                datapath=inner_path,
                data_saveas="xrf_fit_error",
                dataname_saveas="xrf_fit_error_name",
            )

    # TODO: the function 'calculate_roi_sum" should be converted to work with
    #   DASK arrays and HDF5 datasets if needed, otherwise it can be removed.
    #   The function was called at the exit of dialog box 'OutputSetup'
    #   (button 'Output Setup') in the Fit tab. This option is eliminated
    #   from the new interface.
    """
    def calculate_roi_sum(self):
        if self.roi_sum_opt['status'] is True:
            low = int(self.roi_sum_opt['low']*100)
            high = int(self.roi_sum_opt['high']*100)
            logger.info('ROI sum at range ({}, {})'.format(self.roi_sum_opt['low'],
                                                           self.roi_sum_opt['high']))
            sumv = np.sum(self.io_model.data_all[:, :, low:high], axis=2)
            self.result_map['ROI'] = sumv
            # save to hdf again, this is not optimal
            self.save2Dmap_to_hdf()
    """

    def get_latest_single_pixel_fitting_data(self):
        r"""
        Returns the latest results of single pixel fitting. The returned results include
        the name of the scaler (None if no scaler is selected) and the dictionary of
        the computed XRF maps for selected emission lines.
        """

        # Find the selected scaler name. Scaler is None if not scaler is selected.
        scaler_name = None
        if self.scaler_index > 0:
            scaler_name = self.scaler_keys[self.scaler_index - 1]

        # Result map. If 'None', then no fitting results exists.
        #   Single pixel fitting must be run
        result_map = self.result_map.copy()

        return result_map, scaler_name

    def get_selected_fitted_map_data(self):
        """
        Returns fitting data selected in XRF Maps tab.

        Returned channel name should be 'sum', 'det1', 'det2' etc. If the selected
        dataset contains maps computed based on ROIs or only positions and scalers,
        then returned channel name is "".
        """
        image_title = self.img_title
        plotted_dict = self.dict_to_plot.copy()

        # Recover channel name
        channel_name = ""  # No name (dictionary with position or scaler data is selected)
        if image_title.endswith("_fit"):
            m = re.search(r"_det\d+_fit$", image_title)
            if m:
                channel_name = m.group(0).split("_")[1].lower()
            else:
                channel_name = "sum"

        # Find the selected scaler name. Scaler is None if not scaler is selected.
        scaler_name = None
        if self.scaler_index > 0:
            scaler_name = self.scaler_keys[self.scaler_index - 1]

        return plotted_dict, channel_name, scaler_name

    def output_2Dimage(
        self,
        *,
        results_path=None,
        dataset_name=None,
        scaler_name=None,
        interpolate_on=None,
        quant_norm_on=None,
        param_quant_analysis=None,
        file_format="tiff",
    ):
        """Read data from h5 file and save them into either tiff or txt.

        Parameters
        ----------
        results_path: str
            path to the root directory where data is to be saved. Data will be saved
            in subdirectories inside this directory.
        dataset_name: str
            name of the dataset (key in the dictionary `self.io_model.img_dict`
        scaler_name: str
            name of the scaler, must exist in `self.io_model.img_dict`
        interpolate_on: bool
            turns interpolation of images to uniform grid ON or OFF
        quant_norm_on: bool
            turns quantitative normalization of images on or off
        param_quant_analysis: ParamQuantitativeAnalysis
            class that contains methods for quantitative normalization
        file_format: str
            output file format, supported values: "tiff" or "txt"
        """
        supported_file_formats = ("tiff", "txt")
        file_format = file_format.lower()
        if file_format not in supported_file_formats:
            raise ValueError(
                f"The value 'file_format={file_format}' is not supported. "
                f"Supported values: {supported_file_formats} "
            )

        if file_format == "tiff":
            dir_prefix = "output_tiff_"
        else:
            dir_prefix = "output_txt_"

        dataset_dict = self.io_model.img_dict[dataset_name]

        _post_name_folder = "_".join(self.data_title.split("_")[:-1])
        output_n = dir_prefix + _post_name_folder
        output_dir = os.path.join(results_path, output_n)

        # self.io_model.img_dict contains ALL loaded datasets, including a separate "positions" dataset
        if "positions" in self.io_model.img_dict:
            positions_dict = self.io_model.img_dict["positions"]
        else:
            positions_dict = {}

        if scaler_name is not None:
            logger.info(f"*** NORMALIZED data is saved. Scaler: '{scaler_name}' ***")

        # Scalers are located in a separate dataset in 'img_dict'. They are also referenced
        #   in each '_fit' dataset (and in the selected dataset 'self.dict_to_plot')
        #   The list of scaler names is used to avoid attaching the detector channel name
        #   to file names that contain scaler data (scalers typically do not depend on
        #   the selection of detector channels.
        scaler_dsets = [_ for _ in self.io_model.img_dict.keys() if re.search(r"_scaler$", _)]
        if scaler_dsets:
            scaler_name_list = list(self.io_model.img_dict[scaler_dsets[0]].keys())
        else:
            scaler_name_list = None

        output_data(
            output_dir=output_dir,
            interpolate_to_uniform_grid=interpolate_on,
            dataset_name=dataset_name,
            quant_norm=quant_norm_on,
            param_quant_analysis=param_quant_analysis,
            dataset_dict=dataset_dict,
            positions_dict=positions_dict,
            file_format=file_format,
            scaler_name=scaler_name,
            scaler_name_list=scaler_name_list,
        )

    def save_result(self, fname=None):
        """
        Save fitting results.

        Parameters
        ----------
        fname : str, optional
            name of output file
        """
        if not fname:
            fname = self.data_title + "_out.txt"
        filepath = os.path.join(self.result_folder, fname)

        area_list = []
        for v in list(self.fit_result.params.keys()):
            if "ka1_area" in v or "la1_area" in v or "ma1_area" in v or "amplitude" in v:
                area_list.append(v)
        try:
            with open(filepath, "w") as myfile:
                myfile.write("\n {:<10} \t {} \t {}".format("name", "summed area", "error in %"))
                for k, v in self.comps.items():
                    if k == "background":
                        continue
                    for name in area_list:
                        if k.lower() in name.lower():
                            std_error = self.fit_result.params[name].stderr
                            if std_error is None:
                                # Do not print 'std_error' if it is not computed by lmfit
                                errorv_s = ""
                            else:
                                errorv = std_error / (self.fit_result.params[name].value + 1e-8)
                                errorv *= 100
                                errorv = np.round(errorv, 3)
                                errorv_s = f"{errorv}%"
                            myfile.write("\n {:<10} \t {} \t {}".format(k, np.round(np.sum(v), 3), errorv_s))
                myfile.write("\n\n")

                # Print the report from lmfit
                # Remove strings (about 50%) on the variables that stayed at initial value
                report = lmfit.fit_report(self.fit_result, sort_pars=True)
                report = report.split("\n")
                report = [s for s in report if "at initial value" not in s and "##" not in s]
                report = "\n".join(report)
                myfile.write(report)

                logger.warning("Results are saved to {}".format(filepath))
        except FileNotFoundError:
            print("Summed spectrum fitting results are not saved.")

    def update_element_info(self):
        # need to clean list first, in order to refresh the list in GUI
        self.selected_index = 0
        self.elementinfo_list = []

        logger.info("The full list for fitting is {}".format(self.param_model.element_list))


def combine_lines(components, element_list, background):
    """
    Combine results for different lines of the same element.
    And also add pileup, userpeak, background, compton and elastic.

    Parameters
    ----------
    components : dict
        output results from lmfit
    element_list : list
        list of elemental lines
    background : array
        background calculated in given range

    Returns
    -------
    dict :
        combined results for elements and other related peaks.
    """
    new_components = {}
    for e in element_list:
        if len(e) <= 4:
            e_temp = e.split("_")[0]
            intensity = 0
            for k, v in components.items():
                if (e_temp in k) and (e not in k):
                    intensity += v
            new_components[e] = intensity
        elif "user" in e.lower():
            for k, v in components.items():
                if e in k:
                    new_components[e] = v
        else:
            comp_name = "pileup_" + e.replace("-", "_") + "_"  # change Si_K-Si_K to Si_K_Si_K
            new_components[e] = components[comp_name]

    # add background and elastic
    new_components["background"] = background
    new_components["compton"] = components["compton"]
    new_components["elastic"] = components["elastic_"]
    return new_components


def extract_strategy(param, name):
    """
    Extract given strategy from param dict.

    Parameters
    ----------
    param : dict
        saving all parameters
    name : str
        strategy name

    Returns
    -------
    dict :
        with given strategy as value
    """
    param_new = copy.deepcopy(param)
    return {k: v[name] for k, v in param_new.items() if k != "non_fitting_values"}


def define_param_bound_type(param, strategy_list=["adjust_element2, adjust_element3"], b_type="fixed"):
    param_new = copy.deepcopy(param)
    for k, v in param_new.items():
        for data in strategy_list:
            if data in list(v.keys()):
                param_new[k][data] = b_type
    return param_new


def extract_result(data, element):
    """
    Extract fitting result returned from fitting of multi files.

    Parameters
    ----------
    data : list
        list of dict
    element : str
        elemental line
    """
    data_map = []
    for v in data:
        data_map.append(v[element])
    return np.array(data_map)


def bin_data_pixel(data, nearest_n=4):
    """
    Bin 3D data according to number of pixels defined in dim 1 and 2.

    Parameters
    ----------
    data : 3D array
        exp data with energy channel in 3rd dim.
    nearest_n : int, optional
        define how many pixels to be considered.
    """
    new_data = np.array(data)
    d_shape = data.shape

    # if nearest_n == 4:
    #     for i in [-1, 1]:
    #         new_data[1:-1, 1:-1, :] += data[1+i:d_shape[0]-1+i, 1:d_shape[1]-1, :]
    #     for j in [-1, 1]:
    #         new_data[1:-1, 1:-1, :] += data[1:d_shape[0]-1, 1+j:d_shape[1]-1+j, :]

    if nearest_n == 4:
        for i in np.arange(d_shape[0] - 1):
            for j in np.arange(d_shape[1] - 1):
                new_data[i, j, :] += new_data[i + 1, j, :] + new_data[i, j + 1, :] + new_data[i + 1, j + 1, :]
        new_data[:-1, :-1, :] /= nearest_n

    if nearest_n == 9:
        for i in [-1, 0, 1]:
            for j in [-1, 0, 1]:
                new_data[1:-1, 1:-1, :] += data[1 + i : d_shape[0] - 1 + i, 1 + j : d_shape[1] - 1 + j, :]

        new_data[1:-1, 1:-1, :] /= nearest_n

    return new_data


def bin_data_spacial(data, bin_size=4):
    """
    Bin 2D/3D data based on first and second dim, i.e., 2 by 2 window, or 4 by 4.

    Parameters
    ----------
    data : array
        2D or 3D dataset
    bin_size : int
        window size. 2 means 2 by 2 window
    """
    if bin_size <= 1:
        return data

    data = np.asarray(data)
    if data.ndim == 2:
        d_shape = np.array([data.shape[0], data.shape[1]]) / bin_size

        data_new = np.zeros([d_shape[0], d_shape[1]])
        for i in np.arange(d_shape[0]):
            for j in np.arange(d_shape[1]):
                data_new[i, j] = np.sum(
                    data[i * bin_size : i * bin_size + bin_size, j * bin_size : j * bin_size + bin_size]
                )
    elif data.ndim == 3:
        d_shape = np.array([data.shape[0], data.shape[1]]) / bin_size

        data_new = np.zeros([d_shape[0], d_shape[1], data.shape[2]])
        for i in np.arange(d_shape[0]):
            for j in np.arange(d_shape[1]):
                data_new[i, j, :] = np.sum(
                    data[i * bin_size : i * bin_size + bin_size, j * bin_size : j * bin_size + bin_size, :],
                    axis=(0, 1),
                )
    return data_new


def conv_expdata_energy(data, width=2):
    """
    Do convolution on the 3rd axis, energy axis.
    Paremeters
    ----------
    data : 3D array
        exp spectrum
    width : int, optional
        width of the convolution function.
    Returns
    -------
    array :
        after convolution
    """
    data_new = np.array(data)
    if width == 2:
        conv_f = [1.0 / 2, 1.0 / 2]
    if width == 3:
        conv_f = [1.0 / 3, 1.0 / 3, 1.0 / 3]
    for i in np.arange(data.shape[0]):
        for j in np.arange(data.shape[1]):
            data_new[i, j, :] = np.convolve(data_new[i, j, :], conv_f, mode="same")

    return data_new


def bin_data_energy2D(data, bin_step=2, axis_v=0, sum_data=False):
    """
    Bin data based on given dim, i.e., a dim for energy spectrum.
    Return a copy of the data. Currently only binning along first
    dim is implemented.

    Parameters
    ----------
    data : 2D array
    bin_step : int, optional
        size to bin the data
    axis_v : int, optional
        along which dir to bin data.
    sum_data : bool, optional
        sum data from each bin or not

    Returns
    -------
    binned data with reduced dim of the previous size.
    """
    if bin_step == 1:
        return data

    data = np.array(data)
    if axis_v == 0:
        if bin_step == 2:
            new_len = data.shape[0] / bin_step
            m1 = data[::2, :]
            m2 = data[1::2, :]
            if sum_data is True:
                return (m1[:new_len, :] + m2[:new_len, :]) / bin_step
            else:
                return m1[:new_len, :]
        elif bin_step == 3:
            new_len = data.shape[0] / bin_step
            m1 = data[::3, :]
            m2 = data[1::3, :]
            m3 = data[2::3, :]
            if sum_data is True:
                return (m1[:new_len, :] + m2[:new_len, :] + m3[:new_len, :]) / bin_step
            else:
                return m1[:new_len, :]
        elif bin_step == 4:
            new_len = data.shape[0] / bin_step
            m1 = data[::4, :]
            m2 = data[1::4, :]
            m3 = data[2::4, :]
            m4 = data[3::4, :]
            if sum_data is True:
                return (m1[:new_len, :] + m2[:new_len, :] + m3[:new_len, :] + m4[:new_len, :]) / bin_step
            else:
                return m1[:new_len, :]


def bin_data_energy3D(data, bin_step=2, sum_data=False):
    """
    Bin 3D data along 3rd axis, i.e., a dim for energy spectrum.
    Return a copy of the data.

    Parameters
    ----------
    data : 3D array
    sum_data : bool, optional
        sum data from each bin or not

    Returns
    -------
    binned data with reduced dim of the previous size.
    """
    if bin_step == 1:
        return data
    data = np.array(data)
    if bin_step == 2:
        new_len = data.shape[2] / 2
        new_data1 = data[:, :, ::2]
        new_data2 = data[:, :, 1::2]
        if sum_data is True:
            return (new_data1[:, :, :new_len] + new_data2[:, :, :new_len]) / bin_step
        else:
            return new_data1[:, :, :new_len]
    elif bin_step == 3:
        new_len = data.shape[2] / 3
        new_data1 = data[:, :, ::3]
        new_data2 = data[:, :, 1::3]
        new_data3 = data[:, :, 2::3]
        if sum_data is True:
            return (new_data1[:, :, :new_len] + new_data2[:, :, :new_len] + new_data3[:, :, :new_len]) / bin_step
        else:
            return new_data1[:, :, :new_len]
    elif bin_step == 4:
        new_len = data.shape[2] / 4
        new_data1 = data[:, :, ::4]
        new_data2 = data[:, :, 1::4]
        new_data3 = data[:, :, 2::4]
        new_data4 = data[:, :, 3::4]
        if sum_data is True:
            return (
                new_data1[:, :, :new_len]
                + new_data2[:, :, :new_len]
                + new_data3[:, :, :new_len]
                + new_data4[:, :, :new_len]
            ) / bin_step
        else:
            return new_data1[:, :, :new_len]


def cal_r2(y, y_cal):
    """
    Calculate r2 statistics.
    Parameters
    ----------
    y : array
        exp data
    y_cal : array
        fitted data
    Returns
    -------
    float
    """
    sse = np.sum((y - y_cal) ** 2)
    sst = np.sum((y - np.mean(y)) ** 2)
    return 1 - sse / sst


def calculate_area(e_select, matv, results, param, first_peak_area=False):
    """
    Parameters
    ----------
    e_select : list
        elements
    matv : 2D array
        matrix constains elemental profile as columns
    results : 3D array
        x, y positions, and each element's weight on third dim
    param : dict
        parameters of fitting
    first_peak_area : Bool, optional
        get overall peak area or only the first peak area, such as Ar_Ka1

    Returns
    -------
    dict :
        dict of each 2D elemental distribution
    """
    total_list = e_select + ["snip_bkg", "r_factor", "sel_cnt", "total_cnt"]
    mat_sum = np.sum(matv, axis=0)

    if len(total_list) != results.shape[2]:
        logger.error(
            f"The number of features in the list ({len(total_list)} "
            f"is not equal to the number of features in the 'results' array "
            f"({results.shape[2]}). This issue needs to be investigated,"
            f"since some of the generated XRF maps may be invalid."
        )

    result_map = dict()
    for i, eline in enumerate(total_list):
        if i < len(e_select):
            # We are processing the component due to emission line (and may be
            #   additional constant spectrum representing background)
            if first_peak_area is not True:
                result_map.update({total_list[i]: results[:, :, i] * mat_sum[i]})
            else:
                if total_list[i] not in K_LINE + L_LINE + M_LINE:
                    ratio_v = 1
                else:
                    ratio_v = get_branching_ratio(total_list[i], param["coherent_sct_energy"]["value"])
                result_map.update({total_list[i]: results[:, :, i] * mat_sum[i] * ratio_v})
        else:
            # We are just copying additional computed data
            result_map.update({eline: results[:, :, i]})

    return result_map


def save_fitted_fig(
    x_v, matv, results, p1, p2, data_all, data_sel_indices, param_dict, result_folder, use_snip=False
):
    """
    Save single pixel fitting results to figs.
    `data_all` can be numpy array, Dask array or RawHDF5Dataset.
    """
    logger.info(f"Saving plots of the fitted data to file. Selection: {tuple(p1)} .. {tuple(p2)}")

    # Convert the 'data_all', which can be numpy array, Dask array or
    #   RawHDF5Dataset into Dask array, so that it could be treated uniformly
    data_all_dask, file_obj = prepare_xrf_map(data_all)
    # Selection (indices of the processed interval) of `data_all` along axis 2
    d_start, d_stop = data_sel_indices
    # 'file_obj' must remain alive until the function exits

    low_limit_v = 0.5

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.set_xlabel("Energy [keV]")
    ax.set_ylabel("Counts")
    max_v = da.max(data_all_dask[p1[0] : p2[0], p1[1] : p2[1], d_start:d_stop]).compute()

    fitted_sum = None
    for m in range(p1[0], p2[0]):
        for n in range(p1[1], p2[1]):
            data_y = data_all_dask[m, n, d_start:d_stop].compute()

            fitted_y = np.sum(matv * results[m, n, :], axis=1)
            if use_snip is True:
                bg = snip_method_numba(
                    data_y,
                    param_dict["e_offset"]["value"],
                    param_dict["e_linear"]["value"],
                    param_dict["e_quadratic"]["value"],
                    width=param_dict["non_fitting_values"]["background_width"],
                )
                fitted_y += bg

            if fitted_sum is None:
                fitted_sum = fitted_y
            else:
                fitted_sum += fitted_y
            ax.cla()
            ax.set_title("Single pixel fitting for point ({}, {})".format(m, n))
            ax.set_xlabel("Energy [keV]")
            ax.set_ylabel("Counts")
            ax.set_ylim(low_limit_v, max_v * 2)

            ax.semilogy(x_v, data_y, label="exp", linestyle="", marker=".")
            ax.semilogy(x_v, fitted_y, label="fit")

            ax.legend()
            output_path = os.path.join(result_folder, "data_out_" + str(m) + "_" + str(n) + ".png")
            plt.savefig(output_path)

    ax.cla()
    sum_y = da.sum(data_all_dask[p1[0] : p2[0], p1[1] : p2[1], d_start:d_stop], axis=(0, 1)).compute()
    ax.set_title("Summed spectrum from point ({},{}) to ({},{})".format(p1[0], p1[1], p2[0], p2[1]))
    ax.set_xlabel("Energy [keV]")
    ax.set_ylabel("Counts")
    ax.set_ylim(low_limit_v, np.max(sum_y) * 2)
    ax.semilogy(x_v, sum_y, label="exp", linestyle="", marker=".")
    ax.semilogy(x_v, fitted_sum, label="fit", color="red")

    ax.legend()
    fit_sum_name = "pixel_sum_" + str(p1[0]) + "-" + str(p1[1]) + "_" + str(p2[0]) + "-" + str(p2[1]) + ".png"
    output_path = os.path.join(result_folder, fit_sum_name)
    plt.savefig(output_path)

    logger.info(f"Fitted data is saved to the directory '{result_folder}'")


def save_fitted_as_movie(
    x_v, matv, results, p1, p2, data_all, param_dict, result_folder, prefix=None, use_snip=False, dpi=150
):
    """
    Create movie to save single pixel fitting resutls.
    """
    total_n = data_all.shape[1] * p2[0]

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.set_aspect("equal")
    ax.set_xlabel("Energy [keV]")
    ax.set_ylabel("Counts")
    max_v = np.max(data_all[p1[0] : p2[0], p1[1] : p2[1], :])
    ax.set_ylim([0, 1.1 * max_v])

    (l1,) = ax.plot(x_v, x_v, label="exp", linestyle="-", marker=".")
    (l2,) = ax.plot(x_v, x_v, label="fit", color="red", linewidth=2)

    # fitted_sum = None
    plist = []
    for v in range(total_n):
        m = v / data_all.shape[1]
        n = v % data_all.shape[1]
        if m >= p1[0] and m <= p2[0] and n >= p1[1] and n <= p2[1]:
            plist.append((m, n))

    def update_img(p_val):
        m = p_val[0]
        n = p_val[1]
        data_y = data_all[m, n, :]

        fitted_y = np.sum(matv * results[m, n, :], axis=1)
        if use_snip is True:
            bg = snip_method_numba(
                data_y,
                param_dict["e_offset"]["value"],
                param_dict["e_linear"]["value"],
                param_dict["e_quadratic"]["value"],
                width=param_dict["non_fitting_values"]["background_width"],
            )
            fitted_y += bg

        ax.set_title("Single pixel fitting for point ({}, {})".format(m, n))
        # ax.set_ylim(low_limit_v, max_v*2)
        l1.set_ydata(data_y)
        l2.set_ydata(fitted_y)
        return l1, l2

    writer = animation.writers["ffmpeg"](fps=30)
    ani = animation.FuncAnimation(fig, update_img, plist)
    if prefix:
        output_file = prefix + "_pixel.mp4"
    else:
        output_file = "fit_pixel.mp4"
    output_p = os.path.join(result_folder, output_file)
    ani.save(output_p, writer=writer, dpi=dpi)


def fit_per_line_nnls(row_num, data, matv, param, use_snip, num_data, num_feature):
    """
    Fit experiment data for a given row using nnls algorithm.

    Parameters
    ----------
    row_num : int
        which row to fit
    data : array
        selected one row of experiment spectrum
    matv : array
        matrix for regression analysis
    param : dict
        fitting parameters
    use_snip : bool
        use snip algorithm to remove background or not
    num_data : int
        number of total data points
    num_feature : int
        number of data features

    Returns
    -------
    array :
        fitting values for all the elements at a given row. Background is
        calculated as a summed value. Also residual is included.
    """
    logger.debug(f"Row number at {row_num}")
    out = []
    bg_sum = 0
    for i in range(data.shape[0]):
        if use_snip is True:
            bg = snip_method_numba(
                data[i, :],
                param["e_offset"]["value"],
                param["e_linear"]["value"],
                param["e_quadratic"]["value"],
                width=param["non_fitting_values"]["background_width"],
            )
            y = data[i, :] - bg
            bg_sum = np.sum(bg)

        else:
            y = data[i, :]

        result, res = nnls_fit(y, matv, weights=None)
        sst = np.sum((y - np.mean(y)) ** 2)
        if not math.isclose(sst, 0, abs_tol=1e-20):
            r2_adjusted = 1 - res / (num_data - num_feature - 1) / (sst / (num_data - 1))
        else:
            # This happens if all elements of 'y' are equal (most likely == 0)
            r2_adjusted = 0
        result = list(result) + [bg_sum, r2_adjusted]
        out.append(result)
    return np.array(out)


def fit_pixel_multiprocess_nnls(exp_data, matv, param, use_snip=False, lambda_reg=0.0):
    """
    Multiprocess fit of experiment data.

    Parameters
    ----------
    exp_data : array
        3D data of experiment spectrum
    matv : array
        matrix for regression analysis
    param : dict
        fitting parameters
    use_snip : bool, optional
        use snip algorithm to remove background or not
    lambda_reg : float, optional
        applied L2 norm regularizaiton if set above zero

    Returns
    -------
    dict :
        fitting values for all the elements
    """
    num_processors_to_use = multiprocessing.cpu_count()

    logger.info("cpu count: {}".format(num_processors_to_use))

    # TODO: find better permanent solution for issue with multiprocessing on Catalina MacOS
    # =======================================================================================
    # The following is a temporary patch for the issue with multiprocessing on Catalina MacOS.
    # The issue is solved by disabling multiprocessing if current OS is Catalina MacOS.
    # The code will change when the better solution is found.

    # Check if the OS is MacOS Catalina (10.15) or older
    disable_multiprocessing = False
    try:
        os_version = platform.mac_ver()[0]
        # 'os_version' is an empty string if the system is not MacOS
        # os_version has format similar to '10.15.3', MacOS Catalina is 10.15
        if os_version and (LooseVersion(os_version) >= LooseVersion("10.15")):
            disable_multiprocessing = True
    except Exception as ex:
        logger.error(f"Error occurred while checking the version of MacOS: {ex}")

    if disable_multiprocessing:
        pool = multiprocessing.pool.ThreadPool(num_processors_to_use)
        logger.warning(
            "Multiprocessing is currently not supported when running PyXRF in MacOS Catalina. "
            "Computations are executed in multithreading mode instead."
        )
    else:
        pool = multiprocessing.Pool(num_processors_to_use)
        logger.info("Computations are executed in multiprocessing mode.")

    n_data, n_feature = matv.shape

    if lambda_reg > 0:
        logger.info("nnls fit with regularization term, lambda={}".format(lambda_reg))
        diag_m = np.diag(np.ones(n_feature)) * np.sqrt(lambda_reg)
        matv = np.concatenate((matv, diag_m), axis=0)
        exp_tmp = np.zeros([exp_data.shape[0], exp_data.shape[1], n_feature])
        exp_data = np.concatenate((exp_data, exp_tmp), axis=2)

    result_pool = [
        pool.apply_async(fit_per_line_nnls, (n, exp_data[n, :, :], matv, param, use_snip, n_data, n_feature))
        for n in range(exp_data.shape[0])
    ]

    results = []
    for r in result_pool:
        results.append(r.get())

    pool.terminate()
    pool.join()

    # chop data back
    if lambda_reg > 0:
        matv = matv[:-n_feature, :]
        exp_data = exp_data[:, :, :-n_feature]

    return np.asarray(results)


def spectrum_nonlinear_fit(pars, x, reg_mat):
    vals = pars.valuesdict()
    return np.sum(vals["a{}".format(i)] * reg_mat[:, i] for i in range(len(vals)))


def residual_nonlinear_fit(pars, x, data=None, reg_mat=None):
    return spectrum_nonlinear_fit(pars, x, reg_mat) - data


def fit_pixel_nonlinear_per_line(row_num, data, x0, param, reg_mat, use_snip):  # c_weight, fit_num, ftol):

    # c_weight = 1
    # fit_num = 100
    # ftol = 1e-3

    elist = param["non_fitting_values"]["element_list"].split(", ")
    elist = [e.strip(" ") for e in elist]

    # LinearModel = lmfit.Model(simple_spectrum_fun_for_nonlinear)
    # for i in np.arange(reg_mat.shape[0]):
    #     LinearModel.set_param_hint('a'+str(i), value=0.1, min=0, vary=True)

    logger.info("Row number at {}".format(row_num))
    out = []
    snip_bg = 0
    for i in range(data.shape[0]):
        if use_snip is True:
            bg = snip_method_numba(
                data[i, :],
                param["e_offset"]["value"],
                param["e_linear"]["value"],
                param["e_quadratic"]["value"],
                width=param["non_fitting_values"]["background_width"],
            )
            y0 = data[i, :] - bg
            snip_bg = np.sum(bg)
        else:
            y0 = data[i, :]

        fit_params = lmfit.Parameters()
        for i in range(reg_mat.shape[1]):
            fit_params.add("a" + str(i), value=1.0, min=0, vary=True)

        result = lmfit.minimize(
            residual_nonlinear_fit, fit_params, args=(x0,), kws={"data": y0, "reg_mat": reg_mat}
        )

        # result = MS.model_fit(x0, y0,
        #                       weights=1/np.sqrt(c_weight+y0),
        #                       maxfev=fit_num,
        #                       xtol=ftol, ftol=ftol, gtol=ftol)
        # namelist = list(result.keys())
        temp = {}
        temp["value"] = [result.params[v].value for v in list(result.params.keys())]
        temp["err"] = [result.params[v].stderr for v in list(result.params.keys())]
        temp["snip_bg"] = snip_bg
        out.append(temp)
    return out


def fit_pixel_multiprocess_nonlinear(data, x, param, reg_mat, use_snip=False):
    """
    Multiprocess fit of experiment data.

    Parameters
    ----------
    data : array
        3D data of experiment spectrum
    param : dict
        fitting parameters

    Returns
    -------
    dict :
        fitting values for all the elements
    """

    num_processors_to_use = multiprocessing.cpu_count()
    logger.info("cpu count: {}".format(num_processors_to_use))
    pool = multiprocessing.Pool(num_processors_to_use)

    # fit_params = lmfit.Parameters()
    # for i in range(reg_mat.shape[1]):
    #     fit_params.add('a'+str(i), value=1.0, min=0, vary=True)

    result_pool = [
        pool.apply_async(fit_pixel_nonlinear_per_line, (n, data[n, :, :], x, param, reg_mat, use_snip))
        for n in range(data.shape[0])
    ]

    results = []
    for r in result_pool:
        results.append(r.get())

    pool.terminate()
    pool.join()

    return results


def get_area_and_error_nonlinear_fit(elist, fit_results, reg_mat):

    mat_sum = np.sum(reg_mat, axis=0)
    area_dict = OrderedDict()
    error_dict = OrderedDict()
    for name in elist:
        area_dict[name] = np.zeros([len(fit_results), len(fit_results[0])])
        error_dict[name] = np.zeros([len(fit_results), len(fit_results[0])])
    # area_dict = OrderedDict({name:np.zeros([len(fit_results), len(fit_results[0])]) for name in elist})
    # error_dict = OrderedDict({name:np.zeros([len(fit_results), len(fit_results[0])]) for name in elist})
    area_dict["snip_bg"] = np.zeros([len(fit_results), len(fit_results[0])])
    weights_mat = np.zeros([len(fit_results), len(fit_results[0]), len(error_dict)])

    for i in range(len(fit_results)):
        for j in range(len(fit_results[0])):
            for m, v in enumerate(area_dict.keys()):
                if v == "snip_bg":
                    area_dict[v][i, j] = fit_results[i][j]["snip_bg"]
                else:
                    area_dict[v][i, j] = fit_results[i][j]["value"][m]
                    error_dict[v][i, j] = fit_results[i][j]["err"][m]
                    weights_mat[i, j, m] = fit_results[i][j]["value"][m]

    for i, v in enumerate(area_dict.keys()):
        if v == "snip_bg":
            continue
        area_dict[v] *= mat_sum[i]
        error_dict[v] *= mat_sum[i]

    return area_dict, error_dict, weights_mat


def single_pixel_fitting_controller(
    input_data,
    parameter,
    incident_energy=None,
    method="nnls",
    pixel_bin=0,
    raise_bg=0,
    comp_elastic_combine=False,
    linear_bg=False,
    use_snip=True,
    bin_energy=1,
    dask_client=None,
):
    """
    Parameters
    ----------
    input_data: array
        3D array of spectrum
    parameter: dict
        parameter for fitting
    incident_energy: float, optional
        incident beam energy in KeV
    method: str, optional
        fitting method, default as nnls
    pixel_bin: int, optional
        bin pixel as 2by2, or 3by3
    raise_bg: int, optional
        add a constant value to each spectrum, better for fitting
    comp_elastic_combine: bool, optional
        combine elastic and compton as one component for fitting
    linear_bg: bool, optional
        use linear background instead of snip
    use_snip: bool, optional
        use snip method to remove background
    bin_energy: int, optional
        bin spectrum with given value
    dask_client: dask.distributed.Client
        Dask client object. If None, then Dask client is created automatically.
        If a batch of files is processed, then creating Dask client and
        passing the reference to it to the processing functions will save
        execution time: `client = Client(processes=True, silence_logs=logging.ERROR)`

    Returns
    -------
    result_map : dict
        of elemental map for given elements
    calculation_info : dict
        dict of fitting information
    """
    param = copy.deepcopy(parameter)
    if incident_energy is not None:
        param["coherent_sct_energy"]["value"] = incident_energy

    n_bin_low, n_bin_high = get_energy_bin_range(
        num_energy_bins=input_data.shape[2],
        low_e=param["non_fitting_values"]["energy_bound_low"]["value"],
        high_e=param["non_fitting_values"]["energy_bound_high"]["value"],
        e_offset=param["e_offset"]["value"],
        e_linear=param["e_linear"]["value"],
    )

    n_bin = np.arange(n_bin_low, n_bin_high)

    # calculate matrix for regression analysis
    elist = param["non_fitting_values"]["element_list"].split(", ")
    elist = [e.strip(" ") for e in elist]
    e_select, matv, e_area = construct_linear_model(n_bin, param, elist)

    # The initial list of elines may contain lines that are not activated for the incident beam
    #   energy. This always happens for at least one line when batches of XRF scans obtained for
    #   the range of beam energies are fitted for the same selection of emission lines to generate
    #   XANES amps. In such experiments, the line of interest is typically not activated at the
    #   lower energies of the band. It is impossible to include non-activated lines in the linear
    #   model. In order to make processing results consistent throughout the batch (contain the
    #   same set of emission lines), the non-activated lines are represented by maps filled with zeros.
    elist_non_activated = list(set(elist) - set(e_select))
    if elist_non_activated:
        logger.warning(
            "Some of the emission lines in the list are not activated: "
            f"{elist_non_activated} at {param['coherent_sct_energy']['value']} keV."
        )

    if comp_elastic_combine is True:
        e_select = e_select[:-1]
        e_select[-1] = "comp_elastic"

        matv_old = np.array(matv)
        matv = matv_old[:, :-1]
        matv[:, -1] += matv_old[:, -1]

    if linear_bg is True:
        e_select.append("const_bkg")

        matv_old = np.array(matv)
        matv = np.ones([matv_old.shape[0], matv_old.shape[1] + 1])
        matv[:, :-1] = matv_old

    logger.info("Matrix used for linear fitting has components: {}".format(e_select))

    def _log_unsupported_option(option):
        logger.warning(
            f"Option '{option}' is enabled. This option is not supported "
            f"and will be ignored. Disable the option to eliminate this warning."
        )

    if raise_bg > 0:
        _log_unsupported_option(f"raise_bg == {raise_bg}")

    # add const background, so nnls works better for values above zero
    # if raise_bg > 0:
    #     exp_data += raise_bg

    if pixel_bin > 1:
        _log_unsupported_option(f"pixel_bin == {pixel_bin}")

    # bin data based on nearest pixels, only two options
    # if pixel_bin in [4, 9]:
    #     logger.info('Bin pixel data with parameter: {}'.format(pixel_bin))
    #     exp_data = bin_data_spacial(exp_data, bin_size=int(np.sqrt(pixel_bin)))
    #     # exp_data = bin_data_pixel(exp_data, nearest_n=pixel_bin)  # return a copy of data

    if bin_energy > 1:
        _log_unsupported_option(f"pixel_bin == {pixel_bin}")
    # bin data based on energy spectrum
    # if bin_energy in [2, 3]:
    #     exp_data = conv_expdata_energy(exp_data, width=bin_energy)

    # make matrix smaller for single pixel fitting
    matv /= input_data.shape[0] * input_data.shape[1]
    # save matrix to analyze collinearity
    # np.save('mat.npy', matv)
    error_map = None

    if method != "nnls":
        logger.warning(f"Fitting using '{method}' is not supported: 'nnls' method will be used instead.")

    logger.info("Fitting method: non-negative least squares")

    snip_param = {
        "e_offset": param["e_offset"]["value"],
        "e_linear": param["e_linear"]["value"],
        "e_quadratic": param["e_quadratic"]["value"],
        "b_width": param["non_fitting_values"]["background_width"],
    }
    results = fit_xrf_map(
        data=input_data,
        data_sel_indices=(n_bin_low, n_bin_high),
        matv=matv,
        snip_param=snip_param,
        use_snip=use_snip,
        chunk_pixels=5000,
        n_chunks_min=4,
        progress_bar=TerminalProgressBar("NNLS fitting"),
        client=dask_client,
    )

    # output area of dict
    result_map = calculate_area(e_select, matv, results, param, first_peak_area=False)

    # Alternative fitting method (nonlinear fit). Very slow and nobody seems to be using it
    # logger.info('Fitting method: nonlinear least squares')
    # matrix_norm = exp_data.shape[0]*exp_data.shape[1]
    # fit_results = fit_pixel_multiprocess_nonlinear(exp_data, x, param, matv/matrix_norm,
    #                                               use_snip=use_snip)
    # result_map, error_map, results = get_area_and_error_nonlinear_fit(e_select,
    #                                                                  fit_results,
    #                                                                  matv/matrix_norm)

    # Generate 'zero' maps for the emission lines that were not activated
    for eline in elist_non_activated:
        result_map[eline] = np.zeros(shape=input_data.shape[0:2])

    calculation_info = dict()
    if error_map is not None:
        calculation_info["error_map"] = error_map

    calculation_info["fit_name"] = e_select
    calculation_info["regression_mat"] = matv
    calculation_info["results"] = results
    calculation_info["fit_range"] = (n_bin_low, n_bin_high)
    calculation_info["energy_axis"] = n_bin
    # Used to be 'exp_data'(selected data), now it is the full dataset,
    #   which can be ndarray, Dask array or RawHDF5Dataset. In order
    #   to get the selected set, 'input_data' must be sliced along axis2
    #   using 'fit_range' values.
    calculation_info["input_data"] = input_data
    calculation_info["data_sel_indices"] = (n_bin_low, n_bin_high)

    return result_map, calculation_info


def get_energy_bin_range(num_energy_bins, low_e, high_e, e_offset, e_linear):
    """
    Find the bin numbers in the range `0 .. num_energy_bins-1` that correspond
    to the selected energy range `low_e` .. `high_e`.

    The range `(n_low:n_high)` includes the bin with energies `low_e` .. `high_e`
    (note that `n_high` is not part of the range).

    Parameters
    ----------
    num_energy_bins : int
        number of energy bins
    low_e : float
        low energy bound, in keV
    high_e : float
        high energy bound, in keV
    e_offset : float
        offset term in energy calibration (energy for bin #0), keV
    e_linear : float
        linear term in energy calibration

    Returns
    -------
    n_low: int
        the number of the bin that corresponds to `low_e`
    n_high: int
        the number of the bin above the one that corresponds to `high_e`
    """

    v_low = (low_e - e_offset) / e_linear
    v_high = (high_e - e_offset) / e_linear + 1

    def _get_index(v):
        return int(np.clip(v, a_min=0, a_max=num_energy_bins))

    n_low, n_high = _get_index(v_low), _get_index(v_high)

    return n_low, n_high


def get_branching_ratio(elemental_line, energy):
    """
    Calculate the ratio of branching ratio, such as ratio of
    branching ratio of Ka1 to sum of br of all K lines.

    It doesn't matter which unit is used for cs calculation,
    as we only want the ratio. So we still use e.cs function.

    Parameters
    ----------
    elemental_line : str
        e.g., 'Mg_K', refers to the K lines of Magnesium
    energy : float
        incident energy in keV

    Returns
    -------
    float :
        calculated ratio
    """

    name, line = elemental_line.split("_")
    e = Element(name)
    transition_lines = TRANSITIONS_LOOKUP[line.upper()]

    sum_v = 0
    for v in transition_lines:
        sum_v += e.cs(energy)[v]
    ratio_v = e.cs(energy)[transition_lines[0]] / sum_v
    return ratio_v


def roi_sum_calculation(dir_path, file_prefix, fileID, element_dict, interpath):
    """
    Parameters
    -----------
    dir_path : str
    file_prefix : str
    fileID : int
    element_dict : dict
        element name with low/high bound
    interpath : str
        path inside hdf5 file to fetch the data

    Returns
    -------
    result : dict
        roi sum for all given elements
    """
    num_str = "{:03d}".format(fileID)
    # logger.info('File number is {}'.format(fileID))
    filename = file_prefix + num_str
    file_path = os.path.join(dir_path, filename)
    with h5py.File(file_path, "r") as f:
        data = f[interpath][:]

    result_map = dict()
    # for v in element_dict.keys():
    #     result_map[v] = np.zeros([datas[0], datas[1]])

    for k, v in element_dict.items():
        result_map[k] = np.sum(data[:, :, v[0] : v[1]], axis=2)

    return result_map


def roi_sum_multi_files(
    dir_path, file_prefix, start_i, end_i, element_dict, interpath="entry/instrument/detector/data"
):
    """
    Fitting for multiple files with Multiprocessing.

    Parameters
    -----------
    dir_path : str
    file_prefix : str
    start_i : int
        start id of given file
    end_i: int
        end id of given file
    element_dict : dict
        dict of element with [low, high] bounds as values
    interpath : str
        path inside hdf5 file to fetch the data

    Returns
    -------
    result : list
        fitting result as list of dict
    """
    num_processors_to_use = multiprocessing.cpu_count()
    logger.info("cpu count: {}".format(num_processors_to_use))
    pool = multiprocessing.Pool(num_processors_to_use)

    result_pool = [
        pool.apply_async(roi_sum_calculation, (dir_path, file_prefix, m, element_dict, interpath))
        for m in range(start_i, end_i + 1)
    ]

    results = []
    for r in result_pool:
        results.append(r.get())

    pool.terminate()
    pool.join()
    return results


def get_cs(elemental_line, eng=12, norm=False, round_n=2):
    """
    Calculate cross section in barns/atom, use e.csb function.

    Parameters
    ----------
    elemental_line: str
        like Pt_L, Si_K
    eng : float
        incident energy in KeV
    norm : bool, optional
        normalized to the primary cs value or not.
    round_n : int
        number of decimal point.
    """
    if "pileup" in elemental_line:
        return "-"
    elif "_K" in elemental_line:
        name_label = "ka1"
        ename = elemental_line.split("_")[0]
    elif "_L" in elemental_line:
        name_label = "la1"
        ename = elemental_line.split("_")[0]
    elif "_M" in elemental_line:
        name_label = "ma1"
        ename = elemental_line.split("_")[0]
    else:
        return "-"

    e = Element(ename)
    sumv = 0
    for line_name in list(e.csb(eng).keys()):
        if name_label[0] in line_name:
            sumv += e.csb(eng)[line_name]
    if norm is True:
        return np.around(sumv / e.csb(eng)[name_label], round_n)
    else:
        return np.around(sumv, round_n)



import os
import yaml
import jsonschema
import numpy as np
import math
import json
import copy
import time as ttime
from .xrf_utils import split_compound_mass, generate_eline_list
from .utils import normalize_data_by_scaler, convert_time_to_nexus_string
import logging

logger = logging.getLogger(__name__)

# ==========================================================================================
#    Functions for operations with YAML files used for keeping descriptions of XRF standards

_xrf_standard_schema = {
    "type": "object",
    "additionalProperties": False,
    "required": ["name", "serial", "description", "compounds"],
    "properties": {
        "name": {"type": "string"},
        "serial": {"type": "string"},
        "description": {"type": "string"},
        "compounds": {
            "type": "object",
            # Chemical formula should always start with a captial letter (Fe2O3)
            "patternProperties": {"^[A-Z][A-Za-z0-9]*$": {"type": "number"}},
            "additionalProperties": False,
            "minProperties": 1,
        },
        "density": {"type": "number"},  # Total density is an optional parameter
    },
}

_xrf_standard_schema_instructions = """
# The file was automatically generated.
#
# Instructions for editing this file:
#
# Description of each standard starts with '- name: ...'. Every following line
#   must be indented by 4 spaces. Each description contains the following items:
#   'name' (name of the standard, arbitrary string), 'serial' (serial number, of
#   the standard, but can be arbitrary string, 'description' (string that contains
#   description of the standard). Those fields may be filled with arbitrary information,
#   best suited to distinguish the standard later. If string consists of only digits
#   (in case of serial number) it must be enclosed in quotes.
#
# The field 'compounds' lists all compounds in the standard. The compounds are
#   presented in the form <compound_formula>: <concentration>.
#   <compound_formula> has to be a valid chemical formula, representing a pure
#   element (C, Fe, Ga, etc.) or compound (Fe2O3, GaAs, etc). Element names
#   must start with a capital letter followed by a lowercase letter (if present).
#   No characters except 'A-Z', 'a-z' and '0-1' are allowed. Lines containing
#   compound specifications must be indented by extra 4 spaces.
#
# The optional field 'density' specifies total density of the sample and used
#   to check integrity of the data (the sum of densities of all compounds
#   must be equaly to 'density' value.
#
# All density values (for compounds and total density) are specified in ug/cm^2
#
# Example (the lines contain extra '#' character, which is not part of YAML file):
#
#-   name: Micromatter 41164
#    serial: '41164'
#    description: CeF3 21.1 / Au 20.6
#    compounds:
#        CeF3: 21.1
#        Au: 20.6
#    density: 41.7
#
# The easiest way to start creating the list of custom standards is to uncomment
#   and edit the following example. To create extra records, duplicate and
#   edit the example or any existing record.

#-    name: Name of the Standard
#     serial: '32654'
#     description: CeF3 21.1 / Au 20.6 (any convenient description)
#     compounds:
#         CeF3: 21.1
#         Au: 20.6

"""


def save_xrf_standard_yaml_file(file_path, standard_data, *, overwrite_existing=False):
    r"""
    Save descriptions of of XRF standards to YAML file

    Parameters
    ----------

    file_path: str
        absolute or relative path to the saved YAML file. If the path does not exist, then
        it is created.

    standard_data: list(dict)
        list of dictionaries, each dictionary is representing the description of one
        XRF standard. Sending ``[]`` will create YAML file, which contains only instructions
        for manual editing of records. Such file can be read by the function
        ``load_xrf_standard_yaml_file``, which returns ``[]``.

    overwrite_existing: bool
        indicates if existing file should be overwritten. Default is False, since
        overwriting of an existing parameter file will lead to loss of data.

    Returns
    -------

        no value is returned

    Raises
    ------

    IOError if the YAML file already exists and ``overwrite_existing`` is not enabled.
    """

    # Make sure that the directory exists
    file_path = os.path.expanduser(file_path)
    file_path = os.path.abspath(file_path)
    flp, _ = os.path.split(file_path)
    os.makedirs(flp, exist_ok=True)

    if not overwrite_existing and os.path.isfile(file_path):
        raise IOError(f"File '{file_path}' already exists")

    s_output = _xrf_standard_schema_instructions
    if standard_data:
        s_output += yaml.dump(standard_data, default_flow_style=False, sort_keys=False, indent=4)
    with open(file_path, "w") as f:
        f.write(s_output)


def load_xrf_standard_yaml_file(file_path, *, schema=_xrf_standard_schema):
    r"""
    Load the list of XRF standard descriptions from YAML file and verify the schema.

    Parameters
    ----------

    file_path: str
        absolute or relative path to YAML file. If file does not exist then IOError is raised.

    schema: dict
        reference to schema used for validation of the descriptions. If ``schema`` is ``None``,
        then validation is disabled (this is not the default behavior).

    Returns
    -------

        list of dictionaries, each dictionary is representing the description of one XRF
        standard samples. Empty dictionary is returned if the file contains no data.

    Raises
    ------

    IOError is raised if the YAML file does not exist.

    jsonschema.ValidationError is raised if schema validation fails.

    RuntimeError if the sum of areal densities of all compounds does not match the
    total density of the sample for at least one sample. The list of all sample
    records for which the data integrity is not confirmed is returned in the
    error message. For records that do not contain 'density' field the integrity
    check is not performed.
    """

    file_path = os.path.expanduser(file_path)
    file_path = os.path.abspath(file_path)

    if not os.path.isfile(file_path):
        raise IOError(f"File '{file_path}' does not exist")

    with open(file_path, "r") as f:
        standard_data = yaml.load(f, Loader=yaml.FullLoader)

    if standard_data is None:
        standard_data = []

    if schema is not None:
        for data in standard_data:
            jsonschema.validate(instance=data, schema=schema)

    # Now check if all densities of compounds sums to total density in every record
    msg = []
    for data in standard_data:
        if "density" in data:
            # The sum of all densities must be equal to total density
            sm = np.sum(list(data["compounds"].values()))
            if not math.isclose(sm, data["density"], abs_tol=1e-6):
                msg.append(f"Record #{data['serial']} ({data['name']}): computed {sm} vs total {data['density']}")
    if msg:
        msg = [f"    {_}" for _ in msg]
        msg = "\n".join(msg)
        msg = "Sum of areal densities does not match total density:\n" + msg
        raise RuntimeError(msg)

    return standard_data


def load_included_xrf_standard_yaml_file():
    r"""
    Load YAML file with descriptions of XRF standards that is part of the
    package.

    Returns
    -------

    List of dictionaries, each dictionary represents description of one XRF standard.

    Raises
    ------

    Exceptions may be raised by ``load_xrf_standard_yaml_file`` function
    """

    # Generate file name (assuming that YAML file is in the same directory)
    file_name = "xrf_quant_standards.yaml"
    file_path = os.path.realpath(__file__)
    file_path, _ = os.path.split(file_path)
    file_path = os.path.join(file_path, file_name)

    return load_xrf_standard_yaml_file(file_path)


def compute_standard_element_densities(compounds):
    r"""
    Computes areal density of each element in the mix of compounds.
    Some compounds in the mix may contain the same elements.

    Parameters
    ----------

    compounds: dict

        dictionary of compound densities: key - compound formula,
        value - density (typically ug/cm^2)

    Returns
    -------

    Dictionary of element densities: key - element name (symbolic),
    value - elmenet density.
    """

    element_densities = {}

    for key, value in compounds.items():
        el_dens = split_compound_mass(key, value)
        for el, dens in el_dens.items():
            if el in element_densities:
                element_densities[el] += dens
            else:
                element_densities[el] = dens

    return element_densities


# ==========================================================================================
#    Functions for operations with JSON files used for keeping quantitative data obtained
#      after processing of XRF standard samples. The data is saved after processing
#      XRF scan of standard samples and later used for quantitative analysis of
#      experimental samples.

_xrf_quant_fluor_schema = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "name",
        "serial",
        "description",
        "element_lines",
        "incident_energy",
        "scaler_name",
        "distance_to_sample",
        "creation_time_local",
        "source_scan_id",
        "source_scan_uid",
    ],
    "properties": {
        # 'name', 'serial' and 'description' (optional) are copied
        #   from the structure used for description of XRF standard samples
        "name": {"type": "string"},
        "serial": {"type": "string"},
        "description": {"type": "string"},
        # The list of element lines. The list is not expected to be comprehensive:
        #   it includes only the lines selected for processing of standard samples.
        "element_lines": {
            "type": "object",
            "additionalProperties": False,
            "minProperties": 1,
            # Symbolic expression representing an element line:
            # Fe - represents all lines, Fe_K - K-lines, Fe_Ka - K alpha lines,
            # Fe_Ka1 - K alpha 1 line. Currently only selections that contain
            # all K, L or M lines is supported.
            "patternProperties": {
                r"^[A-Z][a-z]?(_[KLM]([ab]\d?)?)?$": {
                    "type": "object",
                    "additionalProperties": False,
                    "required": ["density", "fluorescence"],
                    "properties": {"density": {"type": "number"}, "fluorescence": {"type": ["number", "null"]}},
                }
            },
        },
        # Incident energy used in the processing experiment
        "incident_energy": {"type": "number"},
        # Selected channel, expected values are 'sum', 'det1', 'det2', 'det3' etc.
        "detector_channel": {"type": ["string", "null"]},
        # Name of the valid scaler name (specific for data recorded on the beamline
        "scaler_name": {"type": ["string", "null"]},
        # Distance to the sample (number or null)
        "distance_to_sample": {"type": ["number", "null"]},
        # Time of file creation (NEXUS format), optional, null if not set
        "creation_time_local": {"type": ["string", "null"]},
        # Scan ID of the source (scan of the standard), optional, null if not set
        "source_scan_id": {"type": ["integer", "null"]},
        # Scan UID of the source (scan of the standard), optional, null if not set
        "source_scan_uid": {"type": ["string", "null"]},
    },
}


def save_xrf_quant_fluor_json_file(file_path, fluor_data, *, overwrite_existing=False):
    r"""
    Save the results of processing of a scan data for XRF standard sample to a JSON file.
    The saved data will be used later for quantitative analysis of experimental samples.

    Parameters
    ----------

    file_path: str
        absolute or relative path to the saved JSON file. If the path does not exist, then
        it is created.

    fluor_data: dict
        dictionary, which contains the results of processing of a scan of an XRF standard.
        The dictionary should conform to ``_xrf_quantitative_fluorescence_schema``.
        The schema is verified before saving to ensure that the data can be successfully read.

    overwrite_existing: bool
        indicates if existing file should be overwritten. Default is False, since
        overwriting of an existing parameter file will lead to loss of data.

    Returns
    -------

        no value is returned

    Raises
    ------

    IOError if the JSON file already exists and ``overwrite_existing`` is not enabled.

    jsonschema.ValidationError if schema validation fails
    """

    # Note: the schema is fixed (not passed as a parameter). If data format is changed,
    #   then the built-in schema must be changed. The same schema is always used
    #   both for reading and writing of data.
    jsonschema.validate(instance=fluor_data, schema=_xrf_quant_fluor_schema)

    # Make sure that the directory exists
    file_path = os.path.expanduser(file_path)
    file_path = os.path.abspath(file_path)
    flp, _ = os.path.split(file_path)
    os.makedirs(flp, exist_ok=True)

    if not overwrite_existing and os.path.isfile(file_path):
        raise IOError(f"File '{file_path}' already exists")

    s_output = json.dumps(fluor_data, sort_keys=False, indent=4)
    with open(file_path, "w") as f:
        f.write(s_output)


def load_xrf_quant_fluor_json_file(file_path, *, schema=_xrf_quant_fluor_schema):
    r"""
    Load the quantitative data for XRF standard sample from JSON file and verify the schema.

    Parameters
    ----------

    file_path: str
        absolute or relative path to JSON file. If file does not exist then IOError is raised.

    schema: dict
        reference to schema used for validation of the descriptions. If ``schema`` is ``None``,
        then validation is disabled (this is not the default behavior).

    Returns
    -------

        dictionary containing quantitative fluorescence data on XRF sample.

    Raises
    ------

    IOError is raised if the YAML file does not exist.

    jsonschema.ValidationError is raised if schema validation fails.
    """

    file_path = os.path.expanduser(file_path)
    file_path = os.path.abspath(file_path)

    if not os.path.isfile(file_path):
        raise IOError(f"File '{file_path}' does not exist")

    with open(file_path, "r") as f:
        fluor_data = json.load(f)

    if schema is not None:
        jsonschema.validate(instance=fluor_data, schema=schema)

    return fluor_data


def get_quant_fluor_data_dict(quant_param_dict, incident_energy):
    r"""
    Create the dictionary used for storage of data on XRF reference sample. The field
    ``element_lines`` is the dictionary, which stores data on density density of the
    element (in the sample) and fluorescence of the emission line (computed later
    during processing of the reference scan.

    Parameters
    ----------

    quant_param_dict: dict
        Dictionary with the information on reference sample (loaded from YAML configuration
        file). The dictionary should satifsfy the ``_xrf_standard_schema`` schema.

    incident_energy: float
        Incident beam energy

    Returns
    -------

    quant_fluor_data_dict: dict
        Dictionary that contains data on XRF reference sample, including the field
        ``element_lines``, which relates the emission lines (active at ``incident_energy``)
        with respective fluorescence (area under the line spectra) and density of the element.
        The fluorescence is set to None and needs to be computed later. The dictionary
        should satisfy the ``_xrf_quant_fluor_schema`` schema.
    """
    if incident_energy is not None:
        # Make sure that it is 'float', not 'float64', since 'float64' is not supported by 'yaml' package
        incident_energy = float(max(incident_energy, 0))

    quant_fluor_data_dict = {}
    quant_fluor_data_dict["name"] = quant_param_dict["name"]
    quant_fluor_data_dict["serial"] = quant_param_dict["serial"]
    quant_fluor_data_dict["description"] = quant_param_dict["description"]

    # Find the density (mass) of each element in the mis of compounds.
    #   Note, that the sample may contain the same element as a component of multiple compounds.
    element_dict = {}
    for compound, mass in quant_param_dict["compounds"].items():
        # Split compound/compound_density into elements/element_density
        el_and_mass = split_compound_mass(compound, mass)
        for el, ms in el_and_mass.items():
            if el in element_dict:
                element_dict[el] += ms
            else:
                element_dict[el] = ms

    # Create the dictionary of element lines. Fluorescence is unknown at this point,
    #   so it is always None.
    element_lines = {}
    for el, ms in element_dict.items():
        lines = generate_eline_list([el], incident_energy=incident_energy)
        e = {_: {"density": ms, "fluorescence": None} for _ in lines}
        element_lines.update(e)

    quant_fluor_data_dict["element_lines"] = element_lines

    quant_fluor_data_dict["incident_energy"] = incident_energy

    quant_fluor_data_dict["detector_channel"] = None
    quant_fluor_data_dict["scaler_name"] = None
    quant_fluor_data_dict["distance_to_sample"] = None

    quant_fluor_data_dict["creation_time_local"] = None
    quant_fluor_data_dict["source_scan_id"] = None
    quant_fluor_data_dict["source_scan_uid"] = None

    return quant_fluor_data_dict


def fill_quant_fluor_data_dict(quant_fluor_data_dict, *, xrf_map_dict, scaler_name):
    r"""
    Computes average normalized fluorescence values for element lines that are part of
    the reference sample and listed in ``quant_fluor_data_dict["element_lines"]`` and
    present in the ``xrf_map_dict`` and writes the result to
    ``quant_fluor_data_dict["element_lines"][<element_line>]["fluorescence"]``.
    Fluorescence is set to None for the element lines that are not present in ``xrf_map_dict``.
    Element lines that are present in ``xrf_map_dict``, but not part of the reference
    standard, are ignored. If `scaler_name` is one of the keys of ``xrf_map_dict``, then
    fluorescence map is normalized by the scaler before average value is computed. If
    ``scaler_name`` is not one of the keys of ``xrf_map_dict`` or set to None, then
    the average fluorescence is computed without normalization.

    Pixels along the edges of the map are very likely to contain outliers, so the edges
    are not used in computation whenever sufficient data is available (if map contains
    more than 2 pixels along a dimension, the first and the last pixels are not used
    for averaging)

    Parameters
    ----------

    quant_fluor_data_dict: dict
        Dictionary with XRF reference sample data. This dictionary is modified by the function.
        The dictionary must satisfy the '_xrf_quant_fluor_schema' schema.

    xrf_map_dict: dict(array)
        The dictionary of 2D ndarrays, which contain XRF maps for element lines and scalers.
        Dictionary keys are the names of the emission lines (e.g. Fe_K, S_K, Au_M etc.)

    scaler_name: str
        Scaler name. In order for the scaler to be applied, the name must match one of the
        keys of `xrf_map_dict'. If the scaler name is not one of the list keys or set to None,
        then fluorescence is not normalized

    Returns

        No value is returned. The computed fluorescence for the element lines is saved to
        ``quant_fluor_data_dict["element_lines"][<element_line>]["fluorescence"]``

    """

    if not scaler_name:
        logger.warning("No scaler is selected for computing quantitative coefficients. Data is not normalized.")
    elif scaler_name not in xrf_map_dict:
        logger.warning("Scaler '{scaler_name}' is not in XRF map dictionary. Normalization can not be performed.")
        scaler_name = None

    # Clear ALL fluorescence values. Don't touch any other data
    for eline, info in quant_fluor_data_dict["element_lines"].items():
        info["fluorescence"] = None

    # Save the scaler name
    quant_fluor_data_dict["scaler_name"] = scaler_name

    # Compute fluorescence of the emission lines
    eline_list = tuple(quant_fluor_data_dict["element_lines"].keys())
    for eline, map in xrf_map_dict.items():
        if eline in eline_list:
            # Normalize the map if scaler is selected. (Typically scaler IS selected.)
            if scaler_name:
                norm_map = normalize_data_by_scaler(xrf_map_dict[eline], xrf_map_dict[scaler_name])
            else:
                norm_map = xrf_map_dict[eline]
            # Ignore pixels along the edges (those pixels are likely to be outliers that will visibly bias
            #   the mean value in small calibration scans). If scan is smaller that 2 pixels along any
            #   dimension, then all pixels are used, including edges.

            def _get_range(n_elements):
                if n_elements > 2:
                    n_min, n_max = 1, n_elements - 1
                else:
                    n_min, n_max = 0, n_elements
                return n_min, n_max

            ny_min, ny_max = _get_range(norm_map.shape[0])
            nx_min, nx_max = _get_range(norm_map.shape[1])
            mean_fluor = np.mean(norm_map[ny_min:ny_max, nx_min:nx_max])
            # Note: numpy 'float64' is explicitely converted to 'float'
            #     (yaml package does not seem to support 'float64')
            quant_fluor_data_dict["element_lines"][eline]["fluorescence"] = float(mean_fluor)


def prune_quant_fluor_data_dict(quant_fluor_data_dict):
    r"""
    Prunes the fluorescence data dictionary by removing the element lines that are not
    present (fluorescence is None) or have fluorescence <= 0. 'Pruning' is performed before
    saving calibration data, so that only meaningful information is saved.
    The function does not modify the original data structure. Instead it returns the
    copy of the original dictionary with some emission line removed.

    Parameters
    ----------

    quant_fluor_data_dict: dict

        Dictionary with XRF reference sample data. This dictionary is modified by the function.
        The dictionary must satisfy the '_xrf_quant_fluor_schema' schema.

    Returns
    -------

        Copy of ``quant_fluor_data_dict`` with some emission lines removed. Only the emission
        lines that have fluorescence set to valid value are left.
    """
    quant_fluor_data_dict = copy.deepcopy(quant_fluor_data_dict)
    for key, val in quant_fluor_data_dict["element_lines"].copy().items():
        if (val["fluorescence"] is None) or (val["fluorescence"] <= 0):
            del quant_fluor_data_dict["element_lines"][key]

    return quant_fluor_data_dict


def set_quant_fluor_data_dict_optional(quant_fluor_data_dict, *, scan_id=None, scan_uid=None):
    r"""
    Set optional parameters in the existing dictionary with quantitative fluorescence data.
    The parameters include: source_scan_id (if provided), source_scan_uid (if provided),
    and creation time (local time). The function modifies the dictionary ``quant_fluor_data_dict``

    Parameters
    ----------

    quant_fluor_data_dict: dict

        Dictionary with XRF reference sample data. This dictionary is modified by the function.
        The dictionary must satisfy the '_xrf_quant_fluor_schema' schema.

    scan_id: int or str
        Scan ID, must be positive int or a string representing int. If None, then the current
        value of Scan ID is not changed.

    scan_uid: str
        Scan UID. UID has defined format, but it is not checked by the function, so it may
        be any string. If None, then the current value of Scan UID is not changed.

    Returns
    -------

        No value is returned. Instead ``quant_fluor_data_dict`` is modified.
    """
    # Set scan ID (optional parameter)
    if scan_id is not None:
        try:
            # Attempt to convert to 'int' (from str, 'int64' etc)
            scan_id = int(scan_id)
        except Exception:
            raise RuntimeError("Parameter 'scan_id' must be integer or a string representing integer")

        quant_fluor_data_dict["source_scan_id"] = scan_id

    # Set scan UID (optional parameter)
    if scan_uid is not None:
        if not isinstance(scan_uid, str):
            raise RuntimeError("Parameter 'scan_uid' must be a string representing scan UID")

        quant_fluor_data_dict["source_scan_uid"] = scan_uid

    # Set time as well (it may be changed later)
    set_quant_fluor_data_dict_time(quant_fluor_data_dict)


def set_quant_fluor_data_dict_time(quant_fluor_data_dict):
    r"""
    Set creation time to current local time

    Parameters
    ----------

    quant_fluor_data_dict: dict

        Dictionary with XRF reference sample data. This dictionary is modified by the function.
        The dictionary must satisfy the '_xrf_quant_fluor_schema' schema.

    Returns
    -------

        No value is returned. Instead ``quant_fluor_data_dict`` is modified.
    """
    # TODO: Documentation and tests

    # Set creation time (current local time in NEXUS format)
    quant_fluor_data_dict["creation_time_local"] = convert_time_to_nexus_string(ttime.localtime())


# -------------------------------------------------------------------------------------------------


class ParamQuantEstimation:
    r"""
    The class is used for measurement of parameters in the process of estimation of
    quantitative calibration data.

    The methods of the class are designed to be used in the followingsequence:

    # Create object
    pqe = ParamQuantEstimation()
    # Load standards
    pqe.load_standards()
    # Find standard if needed (different options are available)
    st = pqe.find_standard(serial_number, key="serial")
    # Select standard
    pqe.set_selected_standard(st)

    # Generate data dictionary
    pqe.gen_fluorescence_data_dict(incident_energy=12.0)
    # Fill the dictionary using XRF map dictionary (e.g. ``img_dict``) and scaler name (e.g. ``i0``)
    pqe.fill_fluorescence_data_dict(xrf_map_dict=img_dict, scaler_name="i0")
    # Set different (optional but desired) parameters
    pqe.set_detector_channel_in_data_dict(detector_channel="sum")
    pqe.set_optional_parameters(scan_id=12345, scan_uid="some-uid-string")

    # Get preview (if needed) for displaying to users
    preview_str = pqe.get_fluorescence_data_dict_text_preview()

    # Get suggested file name for the parameter file
    fln = pqe.get_suggested_json_fln()
    # Generate full path based on fln
    file_path = .....
    # Save data to file
    pqe.save_fluorescence_data_dict(file_path=file_path)

    At any time, the generated/filled fluorescence data dictionary may be
    accessed as ``self.fluorescence_data_dict``.
    """

    def __init__(self, *, home_dir="~", config_dir=".pyxrf", standards_fln="quantitative_standards.yaml"):
        r"""
        Constructor of the ``ParamQuantEstiomation`` class. In addition to initalization
        of the fields, the constructor checks if the default file with user-defined
        reference standards exists in the PyXRF config directory and creates an empty file
        if it does not exist.

        Parameters
        ----------

        home_dir: str
            HOME directory for PyXRF configuration files. Typically it is ``~``,
            but may be changed to temporary directory to run unit tests

        config_dir: str
            subdirectory in the HOME directory, used to keep PyXRF config files.
            Typical value is ``.pyxrf``.

        standards_fln: str
            name of the file for storing the data on user-defined reference standards.
        """

        custom_path = (os.path.expanduser(home_dir), config_dir, standards_fln)
        self.custom_standards_file_path = os.path.join(*custom_path)

        # If file with user-defined set of reference standards does not exist, create one
        if not os.path.isfile(self.custom_standards_file_path):
            try:
                save_xrf_standard_yaml_file(self.custom_standards_file_path, [])
            except Exception as ex:
                logger.error(f"Failed to create empty file for custom set of quantitative standards: {ex}")

        # Lists of standard descriptions
        self.standards_built_in = None
        self.standards_custom = None

        # Reference to the selected standard description (in custom or built-in list)
        self.standard_selected = None

        # List of emission lines for the selected incident energy
        self.incident_energy = 0.0
        self.emission_line_list = None

        # Dictionary with fluorescence data for the selected standard. Filled as data is processed
        self.fluorescence_data_dict = None

    def load_standards(self):
        r"""
        Load reference standards data (both built-in and user-defined). Must be called before
        attempting to access the reference standards
        """
        self.clear_standards()

        try:
            self.standards_built_in = load_included_xrf_standard_yaml_file()
        except Exception as ex:
            self.standards_built_in = None
            logger.error(f"Failed to load built-in set of quantitative standards: {ex}")

        try:
            self.standards_custom = load_xrf_standard_yaml_file(self.custom_standards_file_path)
        except Exception as ex:
            self.standards_custom = None
            logger.error(f"Failed to load custom set of quantitative standards: {ex}")

    def clear_standards(self):
        r"""
        Clear the lists of reference standards and all data that was computed based on
        loaded reference standards. This function is called before reloading the standards.
        It is recommended that the data is reloaded often (each time the dialog box
        for reference standard selection is opened), because the reference files may be
        edited by the user and the most recent version of the data should be displayed
        in the dialog box
        """
        self.standards_built_in = None
        self.standards_custom = None
        self.standard_selected = None
        self.emission_line_list = None
        self.fluorescence_data_dict = None

    def _find_standard_custom(self, standard, key=None):
        r"""
        Search for standard in user-defined list. For more detailed description
        see docstring for ``find_standard`` function
        """
        standard_ref = None
        if self.standards_custom:
            for st in self.standards_custom:
                if (st == standard) if (key is None) else (st[key] == standard):
                    standard_ref = st
                    break
        return standard_ref

    def _find_standard_built_in(self, standard, key=None):
        r"""
        Search for standard in built-in list. For more detailed description
        see docstring for ``find_standard`` function
        """

        standard_ref = None
        if self.standards_built_in:
            for st in self.standards_built_in:
                if (st == standard) if (key is None) else (st[key] == standard):
                    standard_ref = st
                    break
        return standard_ref

    def find_standard(self, standard, key=None):
        r"""
        Search for the standard in the lists of user-defined or built-in standards.
        Search may be performed by using the complete standard information (dictionary)
        or by one of the keys. The function returns reference to the first matching
        entry in the list or None.

            Examples:
            ``find_standard(st)`` - search for dictionary ``st``
            ``find_standard(st["name"], key="name")`` - search by key ``name``.

        Parameters
        ----------

        standard: dict or other
            The dictionary with standard information (the complete dictionary
            must match one of the list elements) or the value of one of the
            dictionary element (e.g. serial, or name). In the latter case,
            the ``key`` must be specified

        key: str or None
            The name of the dictionary key used for searching. The first
            matching entry is returned.

        Returns
        -------

            Reference to one of the entries of user-defined or built-in lists
            if search is successful or None if the standard was not found.
        """
        if standard is None:
            return None

        standard_ref = self._find_standard_custom(standard, key=key)
        if not standard_ref:
            standard_ref = self._find_standard_built_in(standard, key=key)

        return standard_ref

    def set_selected_standard(self, standard=None):
        r"""
        Set standard ``standard`` as currently selected. If ``standard`` does not
        exist in user-defined or built-in list, then the first entry of the
        user-defined list is set as current. If user-defined list is empty or
        None, then the first entry of the built-in list is set as current.
        If both lists are empty or None, then then nothing is selected.
        If ``standard`` is None, then the first available standard is selected.

        Parameters
        ----------

        standard: dict
            The dictionary holding standard information. The complete dictionary
            is compared with entries of internally stored lists to find the match.

        Returns
        -------
            Returns reference to the selected standard (None if no entry is selected)
        """

        standard_ref = self.find_standard(standard)

        if not standard_ref:
            # Set reference pointing to the first available standard description
            if self.standards_custom:
                self.standard_selected = self.standards_custom[0]
            elif self.standards_built_in:
                self.standard_selected = self.standards_built_in[0]
            else:
                self.standard_selected = None

        else:
            # The reference was found in one of the arrays
            self.standard_selected = standard_ref

        return self.standard_selected

    def is_standard_custom(self, standard):
        r"""
        Returns True if standard ``standard`` is user-defined and False otherwise

        Parameters
        ----------

        standard: dict
            The dictionary holding standard information. The complete dictionary
            is compared with entries of internally stored lists to find the match.
        """
        return bool(self._find_standard_custom(standard))

    def gen_fluorescence_data_dict(self, incident_energy):
        r"""
        Generate fluorescence data dictionary based on the description of the selected
        reference standard (reference standard must be selected using ``set_selected_standard``).

        Parameters
        ----------

        incident_energy: float

            Incident beam energy, keV
        """

        if incident_energy:
            self.incident_energy = incident_energy
        self.incident_energy = float(max(self.incident_energy, 0.0))
        # 'incident_energy' must be 'float', not numpy 'float64' in order to be correctly displayed
        #   in 'yaml' formatted preview ('yaml' package does not support 'float64')

        if incident_energy == 0.0:
            logger.warning("Attempting to compute the list of emission lines with incident energy set to 0")

        self.fluorescence_data_dict = get_quant_fluor_data_dict(self.standard_selected, incident_energy)

    def fill_fluorescence_data_dict(self, *, xrf_map_dict, scaler_name):
        r"""
        Fills the generated fluorescence data dictionary based on XRF map dictionary and scaler name.
        The function resets fluorescence for all emission lines in the dictionary to None and then
        iterates through the list of the emission lines. If the map for the emission line is found
        in the ``xrf_map_dict``, then the fluorescence value is computed and filled. Fluorescence
        is computed as mean value of the respective map (over all pixels). If the scaler with name
        ``scaler_name`` is found in ``xrf_map_dict`` then XRF map is normalized by the scaler before
        computing fluorescence, otherwise no normalization is performed.

        Before this function is called, the standards must be loaded (``load_standards``),
        a standard selected (``set_selected_standard``) and fluorescence data dictionary generated
        (``gen_fluorescence_data_dict``).

        Parameters
        ----------

        xrf_map_dict: dict(array)
            The dictionary of 2D ndarrays, which contain XRF maps for element lines and scalers.
            Dictionary keys are the names of the emission lines (e.g. Fe_K, S_K, Au_M etc.)

        scaler_name: str
            Scaler name. In order for the scaler to be applied, the name must match one of the
            keys of `xrf_map_dict'. If the scaler name is not one of the list keys or set to None,
            then fluorescence is not normalized
        """

        fill_quant_fluor_data_dict(self.fluorescence_data_dict, xrf_map_dict=xrf_map_dict, scaler_name=scaler_name)

    def set_detector_channel_in_data_dict(self, *, detector_channel=None):
        r"""
        Set detector channel (``sum``, ``det1``, ``det2`` etc.) in the fluorescence data dictionary.
        Information on the selected detector channel will be used to ensure that correct
        detector channel is selected when processing data (the channels MUST match).

        Parameters
        ----------

        detector_channel: str or None

            The name of the detector channel (``sum``, ``det1``, ``det2`` etc.)
        """
        self.fluorescence_data_dict["detector_channel"] = detector_channel

    def set_distance_to_sample_in_data_dict(self, *, distance_to_sample=None):
        r"""
        Set distance-to-sample in the fluorescence data dictionary. The value must be
        postive float or ZERO. Distance-to-sample is the distance between the sample
        and the detector that is used to adjust calibration in case the distance
        changes between calibration and experimental scans. If distance-to-sample is
        zero, then no adjustment will be performed. If both calibration and experimental
        scans are performed without moving the detector, the value may be kept zero (or None)

        Parameters
        ----------

        distance_to_sample: float

            Distance-to-sample, may be positive float or 0.0.
        """
        if distance_to_sample is not None:
            distance_to_sample = float(max(distance_to_sample, 0.0))
        self.fluorescence_data_dict["distance_to_sample"] = distance_to_sample

    def set_optional_parameters(self, *, scan_id=None, scan_uid=None):
        r"""
        Set optional parameters in the existing dictionary with quantitative fluorescence data.
        The parameters include: source_scan_id (if provided), source_scan_uid (if provided),
        and creation time (local time). The function modifies the dictionary ``quant_fluor_data_dict``

        Parameters
        ----------

        quant_fluor_data_dict: dict

            Dictionary with XRF reference sample data. This dictionary is modified by the function.
            The dictionary must satisfy the '_xrf_quant_fluor_schema' schema.

        scan_id: int or str
            Scan ID, must be positive int or a string representing int. If None, then the current
            value of Scan ID is not changed.

        scan_uid: str
            Scan UID. UID has defined format, but it is not checked by the function, so it may
            be any string. If None, then the current value of Scan UID is not changed.
        """

        set_quant_fluor_data_dict_optional(self.fluorescence_data_dict, scan_id=scan_id, scan_uid=scan_uid)

    def get_suggested_json_fln(self):
        r"""
        Get suggested file name for the reference standard data. File name
        serial number of the standard, therefore the fluorescence data dict must be filled
        before calling this function
        """
        fln = f"standard_{self.fluorescence_data_dict['serial']}.json"
        return fln

    def get_fluorescence_data_dict_text_preview(self, enable_warnings=True):
        r"""
        Generate preview of fluorescence data dictionary. The preview is a human-readable
        multiline string, which displays the dictionary in YAML format. Warnings are
        added to the beginning of the string if ``scaler_name`` or ``distance_to_sample``
        are not set. Warnings can be disabled if ``enable_warnings`` is set to ``False``.

        Parameters
        ----------

        enable_warnings: bool

            True - warnings are enabled, False - warnings are disabled and not be included
            in the output string

        Returns
        -------
        str, dict(str, str)
            A string that contains the preview of the fluorescence data dictionary
            (calibration data) and the dictionary of warnings. Currently two warnings are
            generated: "scaler_missing" - data is not normalized by a scaler,
            "distance_to_sample" - distance-to-sample is set to zero (missing), so
            calibration data can't be used later if distance-to-sample is changed.
        """
        pruned_dict = prune_quant_fluor_data_dict(self.fluorescence_data_dict)
        # Print preview in YAML format (easier to read)
        s = yaml.dump(pruned_dict, default_flow_style=False, sort_keys=False, indent=4)
        msg_warnings = {}
        if enable_warnings:
            if (pruned_dict["scaler_name"] is None) or (pruned_dict["scaler_name"] == ""):
                msg = "WARNING: Scaler is not selected, data is not normalized."
                msg_warnings["scaler_missing"] = msg
            if (pruned_dict["distance_to_sample"] is None) or (pruned_dict["distance_to_sample"] == 0):
                msg = (
                    "WARNING: Distance-to-sample is set to 0 or None. "
                    "Set it to estimated distance between the detector and the standard sample "
                    "if you expect to it to change in the series of scans. Otherwise "
                    "the respective corrections may not be computed. Ignore if the distance "
                    "stays constant throughout the series of scans."
                )
                msg_warnings["distance_to_sample"] = msg
        return s, msg_warnings

    def save_fluorescence_data_dict(self, file_path, *, overwrite_existing=False):
        r"""
        Save the results of processing of a scan data for XRF standard sample to a JSON file.
        The saved data will be used later for quantitative analysis of experimental samples.
        Before saving, the copy of the original dictionary is createda and inactive emission
        lines (with ZERO fluorescence) are removed from the copy. Original dictionary is not
        modified and the copy is saved to the file.

        Before this function is called, the standards must be loaded (``load_standards``),
        a standard selected (``set_selected_standard``),  fluorescence data dictionary generated
        (``gen_fluorescence_data_dict``) and filled (``fill_fluorescence_data_dict``).

        Parameters
        ----------

        file_path: str
            absolute or relative path to the saved JSON file. If the path does not exist, then
            it is created.

        overwrite_existing: bool
            indicates if existing file should be overwritten. Default is False, since
            overwriting of an existing parameter file will lead to loss of data.

        Returns
        -------

            no value is returned

        Raises
        ------

        IOError if the JSON file already exists and ``overwrite_existing`` is not enabled.

        jsonschema.ValidationError if schema validation fails
        """
        # Set time (in the original dictionary)
        set_quant_fluor_data_dict_time(self.fluorescence_data_dict)
        # The next step creates a copy of the dictionary with removed inactive emission lines
        pruned_dict = prune_quant_fluor_data_dict(self.fluorescence_data_dict)
        save_xrf_quant_fluor_json_file(file_path, pruned_dict, overwrite_existing=overwrite_existing)


class ParamQuantitativeAnalysis:
    r"""
    The class is used to load and manipulate quantitative calibration data and apply quantitative
    normalization to XRF maps. Calibration data is loaded as entries, one calibration entry is
    produced by processing one calibration scan based on a reference standards. Typically calibration
    entry is stored in JSON file, but it can be added to the class directly. Multiple entries
    based on different reference standards may be loaded simultaneously. The same emission line
    may be used in different reference standards and therefore be present in multiple calibration
    entries. The class allows to select the source of calibration data (entry) for each emission
    line.

    The functions should be applied in the following sequence:

    pqa = ParamQuantitativeAnalysis()
    # Load calibration entries
    pqa.load_entry(<file_path1>)
    pqa.load_entry(<file_path2>)
    ...
    # or add calibration entries (the 1st parameter doesn't need to be a file path)
    pqa.add_entry("calibration1", calibration_data=<calib1>)
    pqa.add_entry("calibration2", calibration_data=<calib2>)
    ...
    # Entry may be removed as
    pqa.remove_entry(<file_path1>)
    # Get the list of file paths (or IDs)
    file_paths = pqa.get_file_path_list()
    # Get the number of entries
    n_entries = pqa.get_n_entries()
    # Entry index (index of loaded entry in the internal dictionaries)
    entry_index = pqa.find_entry_index(<file_path>)
    # Get text preview of loaded calibration entry
    text_preview = pqa.get_entry_text_preview(<file_path>)
    # Update the internal state of the object (including the list of available emission lines)
    #   Should be run each time the fields of the object are manually modified.
    pqa.update_emission_line_list()
    # Get the complete set of loaded calibration entries for the given emission line
    #   Note, then the index of the entry in 'eline_info' may be different from the
    #   index of the entry, since 'eline_info' contains only the entries which contain calibration
    #   for the emission line. The list 'eline_info' may be useful for GUI presentation of data.
    eline_info = pqa.get_eline_info_complete(<emission line, e.g. "Cu_K">)
    # Select emission line from the standard (useful when emission line is present in mutliple standards)
    pqa.select_eline(<emission_line>, <file_path>)
    # Check if emission line for particular standard is selected
    pqa.is_eline_selected(<emission_line>, <file_path>)
    # Get calibration data for the emission line (only from the selected entry)
    #   The information is useful at the stage of applying quantitative normalization
    eline_calib = pqa.get_calibrations_selected(<emission_line>)
    # Get calibration data on all emission lines as a list of dictionaries (the function
    #   calls 'get_calibrations_selected' for each emission line in the set)
    all_elines_calib = pqa.get_calibrations_selected
    # Set parameters of the experiment (before processing XRF map)
    pqa.set_experiment_incident_energy(<incident_energy>)
    pqa.set_experiment_detector_channel(<detector_channel>)
    pqa.set_experiment_distance_to_sample(<distance_to_sample>)
    # Process XRF map using selected calibrations and set experiment parameters
    data_normalized = apply_quantitative_normalization(<data_in>, scaler_dict=<scaler_dict>,
                                                       scaler_name_default = <scaler_name_default>,
                                                       data_name = <emission_line>,
                                                       name_not_scalable=["sclr", "sclr2", "time" etc.]):
    """

    def __init__(self):
        r"""
        Constructor. Initialization.
        """

        # List of opened calibration standards
        self.calibration_data = []
        self.calibration_settings = []
        self.active_emission_lines = []

        # Parameters of the experiment for the currently processed dataset:
        #   'experiment_detector_channel', 'experiment_incident_energy', 'experiment_distance_to_sample'.
        #   Set those parameters manually before running 'apply_quantitative_normalization'.

        # Detector channel (values 'sum', 'det1', 'det2' etc.) must match
        #   the channel specified for the quantitative calibration data for the emission line.
        #   (check is performed for each emission line separately, since calibration data
        #   may come from different calibration files depending on user's choice).
        self.experiment_detector_channel = None
        # Incident energy values should be approximately equal
        self.experiment_incident_energy = None
        # Distance to sample. If 0 or None, the respective correction is not applied
        self.experiment_distance_to_sample = None

    def load_entry(self, file_path):
        r"""
        Load QA calibration data from file. The calibration data file must be JSON file which
        satisfies schema ``_xrf_quant_fluor_schema``.

        Data may be loaded from the same file only once.
        If data is loaded from the same file repeatedly, then only one (first) copy is going to be
        stored in the memory.

        Parameters
        ----------

        file_path: str

            Full path to file with QA calibration data.
        """
        # May raise exception if reading the file fails
        calib_data = load_xrf_quant_fluor_json_file(file_path)
        # Data will be added only if file path ``file_path`` was not loaded before. Otherwise
        #   the message will be printed.
        self.add_entry(file_path, calibration_data=calib_data)

    def add_entry(self, file_path, *, calibration_data):
        r"""
        Add QA calibration data to the object. Calibration data is represented as a dictionary
        which satisfies ``_xrf_quant_fluor_schema`` schema. Calibration data may be loaded
        from JSON file (with file name ``file_path``), but it may already exist in memory
        as a dictionary (for example it may be generated or stored by some other module of
        the program). ``file_path`` does not have to represent valid file path, but some
        unique meaningful string, which may be used to identify calibration entries.
        This function DOES NOT LOAD calibration data from file! Use the function
        ``load_entry()`` to load data from file.

        Parameters
        ----------

        file_path: str

            String (perferably meaningful and human readable) used to identify the data entry.
            The value of this parameter is used to distinguish between calibration data
            entries (identify duplicates, search for calibration entries etc.). If data is
            loaded from file, then this parameter is naturally set to full path to the file
            that holds calibration data entry.

        calibration_data: dict

            Calibration data entry: dictionary that satisfies schema ``_xrf_quant_fluor_schema``.

        """

        # Do not load duplicates (distinguished by file name). Different file names may contain
        #   the same data, but this is totally up to the user. Loading duplicates should not
        #   disrupt the processing, since the user may choose the source of calibration data
        #   for each emission line.
        if not any(file_path == _["file_path"] for _ in self.calibration_settings):

            self.calibration_data.append(calibration_data)

            settings = {}
            settings["file_path"] = file_path
            settings["element_lines"] = {}
            for ln in calibration_data["element_lines"]:
                settings["element_lines"][ln] = {}
                # Do not select the emission line
                settings["element_lines"][ln]["selected"] = False

            self.calibration_settings.append(settings)
            # This will also select the emission line if it occurs the first time
            self.update_emission_line_list()
        else:
            logger.info(f"Calibration data file '{file_path}' is already loaded")

    def remove_entry(self, file_path):
        r"""
        Remove QA calibration data entry identified by ``file_path`` from memory
        (from object's internal list). Removes the entry if ``file_path`` is found, otherwise
        prints error message (``logger.error``).

        Parameters
        ----------

            The string used to identify QA calibration data entry. See the docstring for
            ``add_entry`` for more detailed comments.
        """
        n_item = self.find_entry_index(file_path)
        if n_item is not None:
            self.calibration_data.pop(n_item)
            self.calibration_settings.pop(n_item)
            self.update_emission_line_list()
        else:
            raise RuntimeError(f"Calibration data from source '{file_path}' is not found.")

    def get_file_path_list(self):
        r"""
        Returns the list of file paths, which identify the available calibration data entries.
        If calibration data entries are not directly loaded from files ``load_entry``,
        but instead added using ``add_entry``, the strings may not represent actual
        file paths, but instead be any strings used to uniquely identify calibration data entries.

        Returns
        -------

        file_path_list: list(str)

            List of strings that identify QA calibration data entries, may represent paths to
            files in which calibration data entries are stored. Returns empty list if no
            QA calibration data are loaded.
        """
        file_path_list = []
        for v in self.calibration_settings:
            file_path_list.append(v["file_path"])
        return file_path_list

    def get_n_entries(self):
        r"""
        Returns the number of QA calibration data entries.

        Returns
        -------

            The number of available calibration entries. If not calibration data was loaded or added,
            then the returned value is 0.
        """
        if not self.calibration_settings:
            return 0
        else:
            return len(self.calibration_settings)

    def find_entry_index(self, file_path):
        r"""
        Find the index of calibration data entry with the given file path. Returns ``None``
        if ``file_path`` is not found.

        Parameters
        ----------

        file_path: str

            The string used to identify QA calibration data entry. See the docstring for
            ``add_entry`` for more detailed comments.

        """
        n_item = None
        for n, v in enumerate(self.calibration_settings):
            if v["file_path"] == file_path:
                n_item = n
                break
        return n_item

    def get_entry_text_preview(self, file_path):
        r"""
        Returns text (YAML) preview of the dictionary with QA calibration data entry.
        Only original data (loaded from file) is printed.

        Parameters
        ----------

        file_path: str

            The string used to identify QA calibration data entry. See the docstring for
            ``add_entry`` for more detailed comments.

        Returns
        -------

            String wich contains text (YAML) representation of QA calibration data entry.
            Empty string if the entry is not found.
        """
        n_item = self.find_entry_index(file_path)
        if n_item is not None:
            s = yaml.dump(self.calibration_data[n_item], default_flow_style=False, sort_keys=False, indent=4)
        else:
            s = ""
        return s

    def select_eline(self, eline, file_path):
        r"""
        Select emission line in the standard pointed by `file_path` as selected.
        The function will also deselect any the emission line in all other loaded standards.
        Use `get_file_path_list()` to get the list of file paths arranged in the order
        in which standards were loaded.

        It is assumed that `eline` exists in the standard `file_path`.

        Parameters
        ----------
        file_path: str
            The string used to identify QA calibration data entry. See the docstring for
            ``add_entry`` for more detailed comments.
        eline: str
            Emission line name. Must match the emission line names used in calibration data.
            Typical format: ``Fe_K``, ``S_K`` etc.
        """
        n_item = self.find_entry_index(file_path)
        for n, settings in enumerate(self.calibration_settings):
            sel_flag = n == n_item  # Set 'selected' to True only if n == n_item
            if eline in settings["element_lines"]:
                self.calibration_settings[n]["element_lines"][eline]["selected"] = sel_flag
        self.update_emission_line_list()

    def is_eline_selected(self, eline, file_path):
        r"""
        Check if the emission line in the standard `file_path` is selected.
        Use `get_file_path_list()` to get the list of file paths arranged in the order
        in which standards were loaded.

        It is assumed that `eline` exists in the standard `file_path`.

        Parameters
        ----------
        file_path: str
            The string used to identify QA calibration data entry. See the docstring for
            ``add_entry`` for more detailed comments.
        eline: str
            Emission line name. Must match the emission line names used in calibration data.
            Typical format: ``Fe_K``, ``S_K`` etc.

        Returns
        -------
        bool
            `True` if eline is selected, `False` otherwise
        """
        n_item = self.find_entry_index(file_path)
        return self.calibration_settings[n_item]["element_lines"][eline]["selected"]

    def update_emission_line_list(self):
        r"""
        Update the internal list of emission lines. Also check and adjust if necessary
        the selected (assigned) calibration data entry for each emission line.

        This function is called by ``load_entry``, ``add_entry``, ``remove_entry`` and
        ``select_eline`` functions, since the set of emission lines is changed during
        those operations. If manual changes are made to loaded calibration data entries,
        this function has to be called again before any other operation is performed.
        Particularly, it is recommended that the function is called after changing the
        selection of calibration sources for emission lines.
        """

        # The emission lines are arranged in the list in the order in which they appear
        elines_all = set()
        self.active_emission_lines = []  # Clear the list, it is recreated
        for data in self.calibration_data:
            for ln in data["element_lines"].keys():
                if ln not in elines_all:
                    elines_all.add(ln)
                    self.active_emission_lines.append(ln)

        # Make sure that emission line is selected only once
        elines_selected = set()  # We make sure that each emission line is selected
        for settings in self.calibration_settings:
            for k, v in settings["element_lines"].items():
                if v["selected"]:
                    if k in elines_selected:
                        # If eline is in the set, then deselect it (it is selected twice)
                        v["selected"] = False
                    else:
                        # Add to the list
                        elines_selected.add(k)

        elines_not_selected = elines_all - elines_selected

        if elines_not_selected:
            for settings in self.calibration_settings:
                for k, v in settings["element_lines"].items():
                    if k in elines_not_selected:
                        v["selected"] = True
                        elines_not_selected.remove(k)

    def get_eline_info_complete(self, eline):
        r"""
        Collects emission line data from each loaded calibration entry and return it as a list
        of dictionaries. The size of the list determines the number of entries in which the
        emission line is present. If the emission line is not present in any loaded calibration set,
        then empty list is returned.

        The returned representation of complete calibration data is used in GUI interface for
        selection of calibration data sources.

        Parameters
        ----------

        eline: str

            Emission line name. Must match the emission line names used in calibration data.
            Typical format: ``Fe_K``, ``S_K`` etc.

        Returns
        -------

        eline_info: list(dict)

            List of dictionaries that hold calibration data for the given emission line.
            The length of the list is equal to the number of calibration data entries that
            contain information on the emission line ``eline``. Each dictionary contains
            the following fields: ``eline_data`` - reference to ``eline`` part of the
            calibration data entry, ``eline_settings`` - reference to ``eline`` part of
            the calibration settings entry (can be used to change the selection status
            of the entry), ``standard_data`` - reference to the full calibration data entry,
            ``standard_settings`` - reference to the full settings data entry. References
            to the full entries seem redundant, but they can be used to access data,
            which is common to the entry.
        """
        eline_info = []
        for n in range(len(self.calibration_data)):
            try:
                # Include only references to
                data = self.calibration_data[n]["element_lines"][eline]
                settings = self.calibration_settings[n]["element_lines"][eline]
                record = {}
                record["standard_data"] = self.calibration_data[n]
                record["standard_settings"] = self.calibration_settings[n]
                record["eline_data"] = data
                record["eline_settings"] = settings
                eline_info.append(record)
            except Exception:
                pass
        return eline_info

    def get_eline_calibration(self, eline):
        r"""
        Returns calibration data for the emission line ``eline`` from the selected
        QA calibration data entry.

        Parameters
        ----------

        eline: str

            Emission line name. Must match the emission line names used in calibration data.
            Typical format: ``Fe_K``, ``S_K`` etc.

        Returns
        -------

        The dictionary with calibration settings. Keys ``density`` - density of the sample
        material in the standard scan (typically in ug/cm^2), ``fluorescence`` - collected
        fluorescence during calibration experiment due to the emission line ``eline``,
        ``incident_energy`` - incident beam energy (in keV) used during the calibration scan,
        ``detector_channel`` - name of the detector channel used in calibration experiment
        (``sum``, ``det1``, ``det2`` etc), ``scaler_name`` - name of the scaler used in
        calibration experiment (valid name from the scan dataset), ``distance_to_sample`` -
        distance from the sample to the detector in calibration scan (units are selected by
        the user, only the ratio between the distances in calibration and experimental scans
        matter, if distance-to-sample is 0, then no correction is applied)
        """
        for n in range(len(self.calibration_data)):
            if (eline in self.calibration_data[n]["element_lines"]) and (
                eline in self.calibration_settings[n]["element_lines"]
            ):
                data = self.calibration_data[n]["element_lines"][eline]
                settings = self.calibration_settings[n]["element_lines"][eline]
                if settings["selected"]:
                    e_info = {
                        "density": data["density"],
                        "fluorescence": data["fluorescence"],
                        "incident_energy": self.calibration_data[n]["incident_energy"],
                        "detector_channel": self.calibration_data[n]["detector_channel"],
                        "scaler_name": self.calibration_data[n]["scaler_name"],
                        "distance_to_sample": self.calibration_data[n]["distance_to_sample"],
                    }
                    return e_info
        return None

    def get_calibrations_selected(self):
        r"""
        Returns the dictionary of calibration data from selected data entries for each
        available emission line: keys - emission line names (``Fe_K``, ``S_K`` etc.),
        values - calibration data for the emission line returned by the function
        ``get_eline_calibration``. See docstring for ``get_eline_calibration`` for
        additional information.

        Returns
        -------

        info: dict(dict)

            Dictionary of dictionaries, which contains selected calibration data for each available
            emission line.
        """
        info = {}
        for eline in self.active_emission_lines:
            e_info = self.get_eline_calibration(eline)
            if e_info:
                info[eline] = e_info
        return info

    def set_experiment_detector_channel(self, detector_channel):
        r"""
        Set detector channel name for the currently processed experimental dataset. The
        channel name is used by the function ``apply_quantitative_normalization``.

        The channel names used in calibration and experimental scans must match.

        Parameters
        ----------

        detector_channel: str

            Name of the detector channel (``sum``, ``det1``, ``det2`` etc)
        """
        self.experiment_detector_channel = detector_channel

    def set_experiment_incident_energy(self, incident_energy):
        r"""
        Set incident energy value for the currently processed experimental dataset. The
        incident energy value is used by the function ``apply_quantitative_normalization``.

        The incident energy value used in calibration and experimental scans must match.

        Parameters
        ----------

        incident_energy: float

            The value of the incident energy in keV.
        """
        self.experiment_incident_energy = incident_energy

    def set_experiment_distance_to_sample(self, distance_to_sample):
        r"""
        Set distance-to-sample value for the currently processed experimental dataset. The
        distance-to-sample value is used by the function ``apply_quantitative_normalization``
        to compute the correction coefficient in case the calibration and experimental scans
        are performed at different distances.

        Parameters
        ----------

        distance-to-sample: float

            The distance between the sample and the detector. Any convenient units can
            be used. The same units must be used for calibration and experimental scans.
            Only the ratio between distance-to-sample values during the experimental
            and calibration scans is used for computations. If distance-to-sample is set
            to 0 for calibration or experimental scan, then the correction coefficient
            is not computed or applied (it is assumed that the calibration and experimental
            scans are performed with the same distance).
        """
        self.experiment_distance_to_sample = distance_to_sample

    def apply_quantitative_normalization(
        self, data_in, *, scaler_dict, scaler_name_default, data_name, name_not_scalable=None
    ):

        r"""
        Apply quantitative normalization to the experimental XRF map ``data_in``. The quantitative
        normalization is applied if the following conditions are met:

        -- ``data_name`` is not in the list ``name_not_scalable``.

        -- ``data_name`` represents an emission line for which quanitative calibration is available
           in one of the entries.

        -- scaler name specified the selected quantitative calibration entry is available
           in ``scaler_dict``.

        -- detector channel selected for the experimental data processing matches the detector
           channel used during calibration.

        If there is significant difference in incident beam energy set for calibration and
        experimental scans, the normalization is still performed, but warning is printed in
        the terminal.

        If quantitative normalization is not performed, but ``scaler_name_fixed`` is a valid key
        of ``scaler_dict``, then regular normalization by the scaler is performed.

        If normalization can not be performed, then the reference to the input array ``data_in``
        is returned.

        The boolean value ``is_quant_normalization_applied`` is set True only if QUANTITATIVE
        normalization is applied.

        Parameters
        ----------

        data_in: ndarray

            XRF map (shape (ny,nx)) needs to be normalized.

        scaler_dict: dict(key: str, value: ndarray)

            Dictionary of the available scaler data for the experimental scan
            (key: scaler name, value: map of scaler values with shape (ny, nx)).

        scaler_name_default: str or None

            Name of the default scaler that is selected for the normalization of experimental
            data. If ``scaler_name_fixed`` is None or not found in the dictionary ``scaler_dict``,
            then normalization is not applied. If quanitative calibration data is available
            for the emission line ``data_name``, then ``scaler_name_fixed`` is ignored and
            the scaler name that was applied during quantitative calibration is applied.
            If ``scaler_name_fixed`` is ``None``, then normalization is applied only if
            quantitative calibration is available for the emission line ``data_name``.

        data_name: str

            The name of XRF map ``data_in``. This may be emission line name (such as ``Fe_K``, ``S_K``)
            or the name of the scalar or positional data. If ``data_name`` represents an emission line,
            then quantitative calibration may be applied to the data.

        name_not_scalable: list or None

            List of map names that are not supposed to be normalized (some scalers, time, positions
            should not be normalized by the scaler). Normalization is skipped for this data.

        Returns
        -------

        data_arr: ndarray

            Output XRF map (normalized or not normalized). If normalization is not applied, then
            output array is the reference to the input array (``data_arr is data_in == True``).

        is_quant_normalization_applied: bool

            Indicates if quantitative normalization was applied to the input map. If normalization
            by fixed (selected) scaler is applied, then the returned value is False.
        """

        logger.debug(
            f"Starting quantiative normalization with scan parameters:\n"
            f"    Detector channel: '{self.experiment_detector_channel}'\n"
            f"    Distance to sample: {self.experiment_distance_to_sample}\n"
            f"    Incident energy: {self.experiment_incident_energy}"
        )

        is_quant_normalization_applied = False

        # Check input data integrity
        if data_in is None:
            return data_in, is_quant_normalization_applied

        # Data name should not necessarily be emission line name. Some data should not be normalized.
        #   Return input data without change
        if name_not_scalable and (data_name in name_not_scalable):
            return data_in, is_quant_normalization_applied

        e_info = self.get_eline_calibration(data_name)
        # Scaler is not strictly required to perform quanitative calibration, so it is allowed
        #   to run the function without a scaler. The scaler name is None if the standard was
        #   processed without normalization, so the sample data will not be normalized as well.
        #   But the results are expected to be much better if the scaler is used.

        run_quant = False
        if e_info:
            run_quant = True
            e_info_scaler = e_info["scaler_name"]

            if (
                (self.experiment_detector_channel is None)
                or (e_info["detector_channel"] is None)
                or (self.experiment_detector_channel.lower() != e_info["detector_channel"].lower())
            ):
                # Detector channels must match. This is critical.
                logger.error(
                    f"Emission line: {data_name}. Mismatch between channels used "
                    f"for calibration ('{e_info['detector_channel']}') and current experimental "
                    f"('{self.experiment_detector_channel}') scans. "
                    f"Quantitative normalization will not be performed."
                )
                run_quant = False

            if (
                (self.experiment_incident_energy is None)
                or (e_info["incident_energy"] is None)
                or not math.isclose(self.experiment_incident_energy, e_info["incident_energy"], abs_tol=0.001)
            ):
                # Still do normalization, but print the warning (result may be inaccurate)
                logger.warning(
                    f"Emission line {data_name}. Incident energy for standard "
                    f"('{e_info['incident_energy']}') and experimental "
                    f"('{self.experiment_incident_energy}') scans don't match. "
                    f"Quantitative normalization will still be performed, but the results may "
                    f"be inaccurate."
                )

            if (e_info_scaler is not None) and (e_info_scaler not in scaler_dict):
                run_quant = False
                logger.error(
                    f"Emission line {data_name}. Scaler '{e_info_scaler}' is not available "
                    f"for this dataset. Quantitative normalization is skipped"
                )

        if run_quant:
            # Quantitative calibration for the emission line is loaded, so normalization is
            #   performed. The scaler used to obtain calibration is used. Note, that
            #   if the scaler is None, then the function returns data without change
            if e_info_scaler is None:
                scaler = None
            else:
                scaler = scaler_dict[e_info_scaler]
            data_arr = normalize_data_by_scaler(
                data_in=data_in, scaler=scaler, data_name=data_name, name_not_scalable=name_not_scalable
            )
            # Normalization function above returns reference if not transformations were applied
            #   Make a copy in this case.
            if data_arr is data_in:
                data_arr = data_in.copy()
            data_arr *= e_info["density"] / e_info["fluorescence"]
            # If distance to sample is set for calibration data and current scan, then apply correction
            #   If either value is ZERO, then don't perform the correction.
            r1 = e_info["distance_to_sample"]
            r2 = self.experiment_distance_to_sample
            if (
                (r1 is not None)
                and (r2 is not None)
                and (r1 > 0)
                and (r2 > 0)
                and not math.isclose(r1, r2, abs_tol=1e-20)
            ):
                # Element density increase as the distance becomes larger
                #   (fluorescence is reduced as r**2)
                data_arr *= (r2 / r1) ** 2
                logger.info(
                    f"Emission line {data_name}. Correction for distance-to_sample was performed "
                    f"(standard: {r1}, sample: {r2})"
                )
            else:
                logger.info(
                    f"Emission line {data_name}. Correction for distance-to_sample was skipped "
                    f"(standard: {r1}, sample: {r2})"
                )
            is_quant_normalization_applied = True
        else:
            # The following condition also takes care of the case when 'scaler_name_fixed' is None
            if scaler_name_default in scaler_dict:
                data_arr = normalize_data_by_scaler(
                    data_in=data_in,
                    scaler=scaler_dict[scaler_name_default],
                    data_name=data_name,
                    name_not_scalable=name_not_scalable,
                )
            else:
                data_arr = data_in

        return data_arr, is_quant_normalization_applied



import sys
import platform
import argparse
from PyQt5.QtWidgets import QApplication, QStyleFactory
from PyQt5.QtGui import QFontDatabase, QPalette, QColor
from PyQt5.QtCore import Qt

from .gui_support.gpc_class import GlobalProcessingClasses
from .gui_module.main_window import MainWindow

import logging

logger = logging.getLogger("pyxrf")

# if hasattr(Qt, 'AA_EnableHighDpiScaling'):
#     QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

# if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
#     QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


def run():
    """Run the application"""
    # import faulthandler
    # faulthandler.enable()

    parser = argparse.ArgumentParser(prog="pyxrf", description="Command line arguments")
    parser.add_argument(
        "-l",
        "--loglevel",
        default="INFO",
        type=str,
        dest="loglevel",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Logger level. Set to 'DEBUG' in order to see debug information.",
    )
    parser.add_argument(
        "-ct",
        "--color-theme",
        default="DEFAULT",
        type=str,
        dest="color_theme",
        choices=["DEFAULT", "DARK"],
        help="Color theme: DARK theme is only for debugging purposes.",
    )
    args = parser.parse_args()

    # Setup the Logger
    logger.setLevel(args.loglevel)
    formatter = logging.Formatter(fmt="%(asctime)s : %(levelname)s : %(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(args.loglevel)
    logger.addHandler(stream_handler)

    # Suppress warnings except if the program is run in debug mode
    if args.loglevel != "DEBUG":
        sys.tracebacklimit = 0

    gpc = GlobalProcessingClasses()
    gpc.initialize()

    app = QApplication(sys.argv)

    # The default font looks bad on Windows, so one of the following (commonly available)
    #   fonts will be selected in the listed order
    windows_font_selection = ["Verdana", "Microsoft Sans Serif", "Segoe UI"]
    available_font_families = list(QFontDatabase().families())
    selected_font_family = None

    current_os = platform.system()
    if current_os == "Linux":
        style = "Fusion"
    elif current_os == "Windows":
        style = "Fusion"
        # Select font
        for font_family in windows_font_selection:
            if font_family in available_font_families:
                selected_font_family = font_family
                break
    elif current_os == "Darwin":
        style = "Fusion"

    available_styles = list(QStyleFactory().keys())
    if style not in available_styles:
        logger.info(f"Current OS: {current_os}")
        logger.info(f"Style '{style}' is not in the list of available styles {available_styles}.")
    app.setStyle(style)
    app.setApplicationName("PyXRF")
    # app.setStyleSheet('QWidget {font: "Roboto Mono"; font-size: 14px}')
    # app.setStyleSheet('QWidget {font-size: 14px}')

    if args.color_theme == "DARK":
        # Custom palette for Dark Mode
        palette = QPalette()
        palette.setColor(QPalette.Window, QColor(53, 53, 53))
        palette.setColor(QPalette.WindowText, Qt.white)
        palette.setColor(QPalette.Base, QColor(25, 25, 25))
        palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
        palette.setColor(QPalette.ToolTipBase, Qt.white)
        palette.setColor(QPalette.ToolTipText, Qt.white)
        palette.setColor(QPalette.Text, Qt.white)
        palette.setColor(QPalette.Button, QColor(53, 53, 53))
        palette.setColor(QPalette.ButtonText, Qt.white)
        palette.setColor(QPalette.BrightText, Qt.red)
        palette.setColor(QPalette.Link, QColor(42, 130, 218))
        palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
        palette.setColor(QPalette.HighlightedText, Qt.black)
        app.setPalette(palette)

    # Set font
    font = app.font()
    font.setPixelSize(14)
    if selected_font_family:
        logger.info(f"Replacing the default font with '{selected_font_family}'")
        font.setFamily(selected_font_family)
    app.setFont(font)

    # The style sheet replicates the default behavior of QToolTip on Ubuntu.
    #   It may be helpful if some strange color theme is used.
    app.setStyleSheet(
        "QToolTip {color: #000000; background-color: #FFFFCC; "
        "border-style: solid; border-radius: 3px; "
        "border-color: #444444; border-width: 1px;}"
    )

    main_window = MainWindow(gpc=gpc)
    main_window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    run()



import time
import numpy as np

# import logging
try:
    from event_model import compose_run
except ModuleNotFoundError:
    pass

import pyxrf
from pyxrf.api import db, db_analysis


def fitting_result_sender(hdr, result, **kwargs):
    """Transfer fitted results to generator.

    Parameters
    ----------
    hdr : header
    result : dict of 2D array
        fitting results
    kwargs : dict
        fitting parameters, exp summed spectrum, fitted summed spectrum
    """
    start_new = dict(hdr.start)
    param = kwargs.get("param", None)
    if param is not None:
        start_new["param"] = param
    start_new["pyxrf_version"] = pyxrf.__version__
    yield "start", start_new
    # first descriptor
    yield "descriptor", {
        "data_keys": {
            "element_name": {"shape": [1], "dtype": "string", "source": "pyxrf fitting"},
            "map": {"shape": hdr.start.shape, "dtype": "array", "source": "pyxrf fitting"},
        },
        "name": "primary",
    }
    time_when_analysis = time.time()
    for i, (k, v) in enumerate(result.items()):
        data = {"element_name": k, "map": v}
        timestamps = {"element_name": time_when_analysis, "map": time_when_analysis}
        event = {"data": data, "seq_num": i + 1, "timestamps": timestamps}
        yield "event", event
    # second descriptor
    yield "descriptor", {
        "data_keys": {
            "summed_spectrum_experiment": {"shape": [4096], "dtype": "array", "source": "experiment"},
            "summed_spectrum_fitted": {"shape": [4096], "dtype": "array", "source": "pyxrf fitting"},
        },
        "name": "spectrum",
    }
    data = {"summed_spectrum_experiment": kwargs["exp"], "summed_spectrum_fitted": kwargs["fitted"]}
    timestamps = {"summed_spectrum_experiment": time_when_analysis, "summed_spectrum_fitted": time_when_analysis}
    event = {"data": data, "seq_num": 1, "timestamps": timestamps}
    yield "event", event

    yield "stop", {}


class ComposeDataForDB:
    """Compose and validate data to format databroker
    can take.
    """

    def __init__(self):
        pass

    def __call__(self, name, doc):
        return name, getattr(self, name)(doc)

    def start(self, doc):
        metadata = {"raw_uid": doc["uid"], "raw_scan_id": doc["scan_id"], "processor_parameters": doc["param"]}
        self.compose_run_bundle = compose_run(metadata=metadata)
        return self.compose_run_bundle.start_doc

    def datum(self, doc):
        return doc

    def resource(self, doc):
        return doc

    def descriptor(self, doc):
        self.compose_descriptor_bundle = self.compose_run_bundle.compose_descriptor(
            name=doc["name"], data_keys=doc["data_keys"], object_names=None, configuration={}, hints=None
        )
        return self.compose_descriptor_bundle.descriptor_doc

    def event(self, doc):
        event_doc = self.compose_descriptor_bundle.compose_event(
            data=doc["data"], timestamps=doc["timestamps"], seq_num=doc["seq_num"]
        )
        return event_doc

    def stop(self, doc):
        return self.compose_run_bundle.compose_stop()


def save_data_to_db(uid, result, doc, db=db, db_analysis=db_analysis):
    """
    Save fitting result to analysis store.

    Parameters
    ----------
    uid : int
        run id
    result : dict of 2D array
        fitting results
    param : dict
        fitting parameters
    db : databroker, optional
        where exp data is saved
    db : analysis store, optional
        where analysis result is to be saved
    """
    print("saving data to db for {}".format(uid))
    hdr = db[uid]
    gen = fitting_result_sender(hdr, result, **doc)
    name, start_doc = next(gen)

    # assert name == 'start'
    processor = ComposeDataForDB()
    # Push the start_doc through.
    _, processed_doc = processor("start", start_doc)
    # insert to analysis store
    db_analysis.insert("start", processed_doc)

    for name, doc in gen:
        print(name)
        _, processed_doc = processor(name, doc)
        # print(processed_doc)
        processed_doc.pop("id", None)
        # insert to analysis store
        db_analysis.insert(name, processed_doc)


def get_analysis_result(raw_scan_id, db_analysis=db_analysis):
    """
    Get data from analysis store with given raw_scan_id=uid.
    More rich search can be implemented here.

    Parameters
    ----------
    raw_scan_id : int
        scan_id of original scan
    db : analysis store, optional
        where analysis result is to be saved

    Returns
    -------
    header
    """
    res = db_analysis(raw_scan_id=raw_scan_id)
    res = list(res)
    if not len(res):
        print("No analysis result is found for {}.".format(raw_scan_id))
        return
    res = sorted(res, key=lambda x: x.start.time)
    uid = res[-1].start.uid  # only latest result for now, to be modified.
    hdr = db_analysis[uid]
    return hdr


# below to be removed
def simulated_result():
    result = {}
    for i in range(5):
        result[f"data_{i}"] = np.ones([10, 10]) + i
    return result


# res = simulated_result()
# param = {'a' : 1, 'b' : 2}
# doc = {'param': param, 'exp': np.arange(10), 'fitted':np.arange(10)}
# save_data_to_db(54132, res, doc)
